# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from pathlib import Path
from typing import Any
from urllib.parse import parse_qs, urlsplit

from pydantic import BaseModel, ConfigDict, Field, model_validator

from catalystwan.endpoints import APIEndpoints, CustomPayloadType, PreparedPayload, get, post, versions, view
from catalystwan.models.tenant import TenantExport
from catalystwan.utils.session_type import ProviderView, SingleTenantView


class MigrationTokenQueryParams(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")
    migration_id: str = Field(serialization_alias="migrationId", validation_alias="migrationId")

    @model_validator(mode="before")
    @classmethod
    def single_migration_id_required(cls, values: Any):
        migration_id = values.get("migrationId")
        if isinstance(migration_id, list) and len(migration_id) == 1:
            values["migrationId"] = migration_id[0]
        return values


class ExportInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    process_id: str = Field(serialization_alias="processId", validation_alias="processId")


class ImportInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    process_id: str = Field(serialization_alias="processId", validation_alias="processId")
    migration_token_url: str = Field(serialization_alias="migrationTokenURL", validation_alias="migrationTokenURL")

    @property
    def migration_token_query(self) -> str:
        return urlsplit(self.migration_token_url).query

    @property
    def migration_token_query_params(self) -> MigrationTokenQueryParams:
        return MigrationTokenQueryParams.model_validate(parse_qs(self.migration_token_query))


class MigrationInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    process_id: str = Field(serialization_alias="processId", validation_alias="processId")


class MigrationFile(CustomPayloadType):
    def __init__(self, filename: Path):
        self.filename = filename

    def prepared(self) -> PreparedPayload:
        data = open(self.filename, "rb")
        return PreparedPayload(files={"file": (Path(data.name).name, data)})


class TenantMigration(APIEndpoints):
    @view({SingleTenantView, ProviderView})
    @versions(">=20.6")
    @get("/tenantmigration/download/{path}")
    def download_tenant_data(self, path: str = "default.tar.gz") -> bytes:
        ...

    @view({SingleTenantView, ProviderView})
    @versions(">=20.6")
    @post("/tenantmigration/export")
    def export_tenant_data(self, payload: TenantExport) -> ExportInfo:
        ...

    @view({SingleTenantView, ProviderView})
    @versions(">=20.6")
    @get("/tenantmigration/migrationToken")
    def get_migration_token(self, params: MigrationTokenQueryParams) -> str:
        ...

    @view({SingleTenantView, ProviderView})
    @versions(">=20.6,<20.13")
    @post("/tenantmigration/import")
    def import_tenant_data(self, payload: MigrationFile) -> ImportInfo:
        ...

    @view({SingleTenantView, ProviderView})
    @versions(">=20.13")
    @post("/tenantmigration/import/{migration_key}")
    def import_tenant_data_with_key(self, payload: MigrationFile, migration_key: str) -> ImportInfo:
        ...

    @view({SingleTenantView, ProviderView})
    @versions(">=20.6")
    @post("/tenantmigration/networkMigration")
    def migrate_network(self, payload: str) -> MigrationInfo:
        ...

    def retrigger_network_migration(self):
        # GET /tenantmigration/networkMigration
        ...



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from typing import TYPE_CHECKING, List

from catalystwan.dataclasses import (
    OmpAdvertisedRouteData,
    OmpAdvertisedTlocData,
    OmpPeerData,
    OmpReceivedRouteData,
    OmpReceivedTlocData,
    OmpServiceData,
    OmpSummaryData,
)
from catalystwan.utils.creation_tools import create_dataclass

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class OmpAPI:
    """OMP API methods of vManage for get common omp data.

    Attributes:
        session (ManagerSession): logged in API client session

    Usage example:
        # Create session and chose device
        session = create_manager_session(...)
        device = DevicesAPI(session).get().filter(personality = Personality.VSMART)[0]
        # Gets OMP peers for a device
        omp_peers = session.api.omp.get_omp_peers(device.id)
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session

    def __str__(self) -> str:
        return str(self.session)

    def get_omp_peers(self, device_id: str) -> List[OmpPeerData]:
        """Gets OMP peers for a device.

        Args:
            device_id (str): device ID (usually system-ip)

        Returns:
            List[OmpPeerData]: OmpPeerData object
        """
        items = self.session.get_data(f"/dataservice/device/omp/peers?deviceId={device_id}")
        return [create_dataclass(OmpPeerData, item) for item in items]

    def get_advertised_routes(self, device_id: str) -> List[OmpAdvertisedRouteData]:
        """Gets OMP advertised routes data for a device.

        Args:
            device_id (str): device ID (usually system-ip)

        Returns:
            List[OmpAdvertisedRouteData]: OmpAdvertisedRouteData objects
        """
        items = self.session.get_data(f"/dataservice/device/omp/routes/advertised?deviceId={device_id}")
        return [create_dataclass(OmpAdvertisedRouteData, item) for item in items]

    def get_received_routes(self, device_id: str) -> List[OmpReceivedRouteData]:
        """Gets OMP received routes data for a device.

        Args:
            device_id (str): device ID (usually system-ip)

        Returns:
            List[OmpReceivedRouteData]: OmpReceivedRouteData objects
        """
        items = self.session.get_data(f"/dataservice/device/omp/routes/received?deviceId={device_id}")
        return [create_dataclass(OmpReceivedRouteData, item) for item in items]

    def get_advertised_tlocs(self, device_id: str) -> List[OmpAdvertisedTlocData]:
        """Gets OMP advertised TLOCs data for a device.

        Args:
            device_id (str): device_id: device ID (usually system-ip)

        Returns:
            List[OmpAdvertisedTlocData]: OmpAdvertisedTlocData objects
        """
        items = self.session.get_data(f"/dataservice/device/omp/tlocs/advertised?deviceId={device_id}")
        return [create_dataclass(OmpAdvertisedTlocData, item) for item in items]

    def get_received_tlocs(self, device_id: str) -> List[OmpReceivedTlocData]:
        """Gets OMP received TLOCs for a device.

        Args:
            device_id (str): device ID (usually system-ip)

        Returns:
            List[OmpReceivedTlocData]: OmpReceivedTlocData objects
        """
        items = self.session.get_data(f"/dataservice/device/omp/tlocs/received?deviceId={device_id}")
        return [create_dataclass(OmpReceivedTlocData, item) for item in items]

    def get_services(self, device_id: str) -> List[OmpServiceData]:
        """Gets OMP services data for a device.

        Args:
            device_id (str): device ID (usually system-ip)

        Returns:
            List[OmpServiceData]: OmpServiceData objects
        """
        items = self.session.get_data(f"/dataservice/device/omp/services?deviceId={device_id}")
        return [create_dataclass(OmpServiceData, item) for item in items]

    def get_omp_summary(self, device_id: str) -> List[OmpSummaryData]:
        """Gets OMP summaries data for a device.

        Args:
            device_id (str): device ID (usually system-ip)

        Returns:
            List[OmpSummaryData]: OmpSummaryData objects
        """
        items = self.session.get_data(f"/dataservice/device/omp/summary?deviceId={device_id}")
        return [create_dataclass(OmpSummaryData, item) for item in items]



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import copy
from typing import Any, Generic, Iterable, MutableSequence, Type, TypeVar, overload

from pydantic import BaseModel

from catalystwan.exceptions import InvalidOperationError
from catalystwan.utils.creation_tools import AttrsInstance, asdict

T = TypeVar("T")
D = TypeVar("D")


class TypedList(MutableSequence[T], Generic[T]):
    """A list where all elements are of the same data type.

    A homogeneous list provides an efficient and organized way to store data,
    as all elements can be assumed to have the same properties and behaviors.
    Built-in mutable sequence.

    If no argument is provided, the constructor creates a new empty list.
    The argument must be an iterable if specified.

    An implementation is based of pythonic list, instead of creating a new
    Linked List from scratch.
    """

    @overload
    def __init__(self, _type: T) -> None:
        ...

    @overload
    def __init__(self, _type: T, _iterable: Iterable[T], /) -> None:
        ...

    def __init__(self, _type, _iterable=None, /):
        self.data = []
        self._type = _type

        if _iterable is not None:
            for item in _iterable:
                if not isinstance(item, self._type):
                    raise TypeError(f"Expected {self._type.__name__} item type, " f"got {type(item).__name__}.")
                self.data.append(item)

    def __repr__(self) -> str:
        return f"TypedList({self._type.__name__}, {repr(self.data)})"

    def __contains__(self, item: Any) -> bool:
        return item in self.data

    def __len__(self) -> int:
        return len(self.data)

    @overload
    def __getitem__(self, i: int) -> T:
        ...

    @overload
    def __getitem__(self, i: slice) -> TypedList[T]:
        ...

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__class__(self._type, self.data[i])
        else:
            return self.data[i]

    @overload
    def __setitem__(self, i: int, item: T, /) -> None:
        ...

    @overload
    def __setitem__(self, i: slice, item: Iterable[T], /) -> None:
        ...

    def __setitem__(self, i, item):
        if not isinstance(item, self._type):
            raise TypeError(f"Expected {self._type.__name__} item type, " f"got {type(item).__name__}.")
        self.data[i] = item

    @overload
    def __delitem__(self, i: int, /) -> None:
        ...

    @overload
    def __delitem__(self, i: slice, /) -> None:
        ...

    def __delitem__(self, i):
        del self.data[i]

    def __add__(self, __value: Iterable[T]) -> TypedList[T]:
        return TypedList(self._type, self.data + [*__value.__iter__()])

    def __iadd__(self, __value: Iterable[T]) -> TypedList[T]:
        self.data = TypedList(self._type, self.data + [*__value.__iter__()]).data
        return self

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, TypedList):
            if len(self) != len(__o):
                return False
            return all([self.data[i] == __o.data[i] for i in range(len(self))])
        return False

    def append(self, item: T) -> None:
        if not isinstance(item, self._type):
            raise TypeError(f"Expected {self._type.__name__} item type, " f"got {type(item).__name__}.")
        self.data.append(item)

    def insert(self, i: int, item: T) -> None:
        if not isinstance(item, self._type):
            raise TypeError(f"Expected {self._type.__name__} item type, " f"got {type(item).__name__}.")
        self.data.insert(i, item)

    def pop(self, i: int = -1) -> T:
        return self.data.pop(i)

    def remove(self, item: T) -> None:
        self.data.remove(item)

    def clear(self) -> None:
        self.data.clear()

    def count(self, item: T) -> int:
        return self.data.count(item)

    def reverse(self) -> None:
        self.data.reverse()


class DataSequence(TypedList[T], Generic[T]):
    """
    ## Example:

    >>> DataSequence(User, [User(username="User1")])
        DataSequence(User, [
            User(username='User1', password=None, group=[], locale=None, description=None, resource_group=None)
    ])
    """

    @overload
    def __init__(self, _type: Type[T]) -> None:
        ...

    @overload
    def __init__(self, _type: Type[T], _iterable: Iterable[T], /) -> None:
        ...

    def __init__(self, _type, _iterable=None, /):
        if not isinstance(_type, AttrsInstance) and not issubclass(_type, BaseModel):
            raise TypeError(
                f"Expected {AttrsInstance.__name__} or {BaseModel.__name__} item type, got {_type.__name__}."
            )

        super().__init__(_type, _iterable)

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, DataSequence):
            if len(self) != len(__o):
                return False
            return all([self.data[i] == __o.data[i] for i in range(len(self))])
        return False

    def __repr__(self) -> str:
        return f"DataSequence({self._type.__name__}, {repr(self.data)})"

    def __str__(self) -> str:
        pretty_message = ""
        for element in self:
            if issubclass(element.__class__, BaseModel):
                pprint = "\n".join(f"    {attr[0]}: {attr[1]}, " for attr in element.dict().items())  # type: ignore
            else:
                pprint = "\n".join(f"    {attr[0]}: {attr[1]}, " for attr in asdict(element).items())  # type: ignore

            pretty_message += f"\n{element.__class__.__name__}(\n" + pprint + "\n)"  # type: ignore
        return pretty_message

    def __add__(self, __value: Iterable[T]) -> DataSequence[T]:
        return DataSequence(self._type, self.data + [*__value.__iter__()])

    def __iadd__(self, __value: Iterable[T]) -> DataSequence[T]:
        self.data = DataSequence(self._type, self.data + [*__value.__iter__()]).data
        return self

    def __copy__(self) -> DataSequence[T]:
        return DataSequence(self._type, self.data)

    def __deepcopy__(self, memo) -> DataSequence[T]:
        if issubclass(self._type, BaseModel):
            return DataSequence(self._type, [o.model_copy(deep=True) for o in self])  # type: ignore
        else:
            return DataSequence(self._type, [copy.deepcopy(o, memo) for o in self])

    @overload
    def single_or_default(self) -> T:
        ...

    @overload
    def single_or_default(self, default: D) -> D:
        ...

    def single_or_default(self, default=None):
        """Returns the only element of a sequence, or a default value if the sequence is empty.

        ## Example:
        >>> seq = DataSequence(User, [User(username="User1")])
        >>> seq.single_or_default()
        User(username='User1', password=None, group=[], locale=None, description=None, resource_group=None)
        >>> seq = DataSequence(User, [User(username="User1"), User(username="User2")])
        >>> seq.filter(username="User1").single_or_default()
        User(username='User1', password=None, group=[], locale=None, description=None, resource_group=None)

        Args:
            default : The default value to return if the sequence is empty. Defaults to None.

        Raises:
            InvalidOperationError: Raises when there is more than one element in the sequence.

        Returns:
            Union[D, T]: The single element of the input sequence, or default.
        """
        if not self.data:
            return default

        if len(self.data) > 1:
            raise InvalidOperationError("The input sequence contains more than one element.")

        return self.data[0]

    def filter(self, **kwargs) -> DataSequence[T]:
        """Filters a sequence of values based on attributes.

        >>> seq = DataSequence(User, [User(username="User1"), User(username="User2")])
        >>> seq.filter(username="User1")
        DataSequence(User, [
            User(username='User1', password=None, group=[], locale=None, description=None, resource_group=None)
        ])

        Returns:
            DataSequence: Filtered DataSequence.
        """
        annotations = set(kwargs.keys())

        return DataSequence(
            self._type, filter(lambda x: all(getattr(x, a) == kwargs[a] for a in annotations), self.data)
        )

    def first(self) -> T:
        """Returns the first element of a sequence.

        ## Example:
        >>> seq = DataSequence(Device, [Device(hostname="dev-1"), Device(hostname="dev-2"), Device(hostname="dev-3")])
        >>> seq.first()
        Device(hostname="dev-1", personality=Personality.EDGE, ...)

        Raises:
            InvalidOperationError: Raises when there is no elements in the sequence.

        Returns:
            [T]: The single element of the input sequence.
        """

        if len(self.data) < 1:
            raise InvalidOperationError("The input sequence contains no elements.")

        return self.data[0]

    def find(self, **kwargs) -> T:
        """Finds first item in sequence matching values based on attributes.
        Works similarily as filter but assures single element is returned or raises exception.

        >>> seq = DataSequence(User, [User(username="User1"), User(username="User2")])
        >>> seq.find(username="User1")
        User(username='User1', password=None, group=[], locale=None, description=None, resource_group=None)

        Returns:
            [T]: The single element of the input sequence.
        """
        annotations = set(kwargs.keys())
        result = next(filter(lambda x: all(getattr(x, a) == kwargs[a] for a in annotations), self.data), None)
        if result is None:
            raise InvalidOperationError(f"Item matching {kwargs} not found in the input sequence")
        return result



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field, field_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class IPSSignatureListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    generator_id: Global[str] = Field(
        serialization_alias="generatorId", validation_alias="generatorId", description="Range 0 to 4294967295"
    )
    signature_id: Global[str] = Field(
        serialization_alias="signatureId", validation_alias="signatureId", description="Range 0 to 4294967295"
    )

    @field_validator("generator_id")
    @classmethod
    def check_generator_id(cls, generator_id: Global):
        assert 0 <= int(generator_id.value) <= 4294967295
        return generator_id

    @field_validator("signature_id")
    @classmethod
    def check_signature_id(cls, signature_id: Global):
        assert 0 <= int(signature_id.value) <= 4294967295
        return signature_id


class IPSSignatureParcel(_ParcelBase):
    entries: List[IPSSignatureListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_signature(self, signature: str):
        generator_id, signature_id = signature.split(":")
        self.entries.append(
            IPSSignatureListEntry(
                generator_id=as_global(generator_id),
                signature_id=as_global(signature_id),
            )
        )



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import ipaddress
from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator

DEFAULT_TRACKER_THRESHOLD = 300
DEFAULT_TRACKER_INTERVAL = 60
DEFAULT_TRACKER_MULTIPLIER = 3
DEFAULT_INTERFACE_MTU = 1400
DEFAULT_INTERFACE_DPD_INTERVAL = 10
DEFAULT_INTERFACE_DPD_RETRIES = 3
DEFAULT_INTERFACE_IKE_VERSION = 2
DEFAULT_INTERFACE_IKE_REKEY_INTERVAL = 14400
DEFAULT_INTERFACE_IPSEC_REKEY_INTERVAL = 3600
DEFAULT_INTERFACE_IPSEC_REPLAY_WINDOW = 512
DEFAULT_INTERFACE_PAIR_ACTIVE_INTERFACE_WEIGHT = 1
DEFAULT_INTERFACE_PAIR_BACKUP_INTERFACE_WEIGHT = 1
DEFAULT_SIG_VPN_ID = 0
DEFAULT_SERVICE_IDLE_TIME = 0
DEFAULT_SERVICE_REFRESH_TIME = 0


class Application(str, Enum):
    SIG = "sig"


class TunnelSet(str, Enum):
    SECURE_INTERNET_GATEWAY_UMBRELLA = "secure-internet-gateway-umbrella"
    SECURE_INTERNET_GATEWAY_ZSCALER = "secure-internet-gateway-zscaler"


class TunnelDcPreference(str, Enum):
    PRIMARY_DC = "primary-dc"
    SECONDARY_DC = "secondary-dc"


class IkeCiphersuite(str, Enum):
    AES256_CBC_SHA1 = "aes256-cbc-sha1"
    AES256_CBC_SHA2 = "aes256-cbc-sha2"
    AES128_CBC_SHA1 = "aes128-cbc-sha1"
    AES128_CBC_SHA2 = "aes128-cbc-sha2"


class IkeGroup(str, Enum):
    TWO = "2"
    FOURTEEN = "14"
    FIFTEEN = "15"
    SIXTEEN = "16"


class IpsecCiphersuite(str, Enum):
    AES256_CBC_SHA1 = "aes256-cbc-sha1"
    AES256_CBC_SHA384 = "aes256-cbc-sha384"
    AES256_CBC_SHA256 = "aes256-cbc-sha256"
    AES256_CBC_SHA512 = "aes256-cbc-sha512"
    AES256_GCM = "aes256-gcm"
    NULL_SHA1 = "null-sha1"
    NULL_SHA384 = "null-sha384"
    NULL_SHA256 = "null-sha256"
    NULL_SHA512 = "null-sha512"


class PerfectForwardSecrecy(str, Enum):
    GROUP_2 = "group-2"
    GROUP_14 = "group-14"
    GROUP_15 = "group-15"
    GROUP_16 = "group-16"
    NONE = "none"


class Interface(FeatureTemplateValidator):
    if_name: str = Field(json_schema_extra={"vmanage_key": "if-name"})
    auto: bool
    shutdown: bool
    description: Optional[str] = None
    unnumbered: bool = True
    address: Optional[ipaddress.IPv4Interface] = None
    tunnel_source: Optional[ipaddress.IPv4Address] = Field(
        default=None, json_schema_extra={"vmanage_key": "tunnel-source"}
    )
    tunnel_source_interface: Optional[str] = Field(
        default=None, json_schema_extra={"vmanage_key": "tunnel-source-interface"}
    )
    tunnel_route_via: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "tunnel-route-via"})
    tunnel_destination: str = Field(json_schema_extra={"vmanage_key": "tunnel-destination"})
    application: Application = Application.SIG
    tunnel_set: TunnelSet = Field(
        TunnelSet.SECURE_INTERNET_GATEWAY_UMBRELLA, json_schema_extra={"vmanage_key": "tunnel-set"}
    )
    tunnel_dc_preference: TunnelDcPreference = Field(
        TunnelDcPreference.PRIMARY_DC, json_schema_extra={"vmanage_key": "tunnel-dc-preference"}
    )
    tcp_mss_adjust: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "tcp-mss-adjust"})
    mtu: int = DEFAULT_INTERFACE_MTU
    dpd_interval: Optional[int] = Field(
        DEFAULT_INTERFACE_DPD_INTERVAL, json_schema_extra={"vmanage_key": "dpd-interval"}
    )
    dpd_retries: Optional[int] = Field(DEFAULT_INTERFACE_DPD_RETRIES, json_schema_extra={"vmanage_key": "dpd-retries"})
    ike_version: int = Field(DEFAULT_INTERFACE_IKE_VERSION, json_schema_extra={"vmanage_key": "ike-version"})
    pre_shared_secret: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "pre-shared-secret"})
    ike_rekey_interval: Optional[int] = Field(
        DEFAULT_INTERFACE_IKE_REKEY_INTERVAL, json_schema_extra={"vmanage_key": "ike-rekey-interval"}
    )
    ike_ciphersuite: Optional[IkeCiphersuite] = Field(
        IkeCiphersuite.AES256_CBC_SHA1, json_schema_extra={"vmanage_key": "ike-ciphersuite"}
    )
    ike_group: IkeGroup = Field(IkeGroup.FOURTEEN, json_schema_extra={"vmanage_key": "ike-group"})
    pre_shared_key_dynamic: bool = Field(True, json_schema_extra={"vmanage_key": "pre-shared-key-dynamic"})
    ike_local_id: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "ike-local-id"})
    ike_remote_id: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "ike-remote-id"})
    ipsec_rekey_interval: Optional[int] = Field(
        DEFAULT_INTERFACE_IPSEC_REKEY_INTERVAL, json_schema_extra={"vmanage_key": "ipsec-rekey-interval"}
    )
    ipsec_replay_window: Optional[int] = Field(
        DEFAULT_INTERFACE_IPSEC_REPLAY_WINDOW, json_schema_extra={"vmanage_key": "ipsec-replay-window"}
    )
    ipsec_ciphersuite: IpsecCiphersuite = Field(
        IpsecCiphersuite.AES256_GCM, json_schema_extra={"vmanage_key": "ipsec-ciphersuite"}
    )
    perfect_forward_secrecy: PerfectForwardSecrecy = Field(
        PerfectForwardSecrecy.NONE, json_schema_extra={"vmanage_key": "perfect-forward-secrecy"}
    )
    tracker: Optional[bool] = None
    track_enable: Optional[bool] = Field(True, json_schema_extra={"vmanage_key": "track-enable"})
    model_config = ConfigDict(populate_by_name=True)


class SvcType(str, Enum):
    SIG = "sig"


class InterfacePair(FeatureTemplateValidator):
    active_interface: str = Field(json_schema_extra={"vmanage_key": "active-interface"})
    active_interface_weight: int = Field(
        DEFAULT_INTERFACE_PAIR_ACTIVE_INTERFACE_WEIGHT, json_schema_extra={"vmanage_key": "active-interface-weight"}
    )
    backup_interface: Optional[str] = Field("None", json_schema_extra={"vmanage_key": "backup-interface"})
    backup_interface_weight: int = Field(
        DEFAULT_INTERFACE_PAIR_BACKUP_INTERFACE_WEIGHT, json_schema_extra={"vmanage_key": "backup-interface-weight"}
    )
    model_config = ConfigDict(populate_by_name=True)


class DisplayTimeUnit(str, Enum):
    MINUTE = "MINUTE"
    HOUR = "HOUR"
    DAY = "DAY"


class RefreshTimeUnit(str, Enum):
    MINUTE = "MINUTE"
    HOUR = "HOUR"
    DAY = "DAY"


class Service(FeatureTemplateValidator):
    svc_type: SvcType = Field(SvcType.SIG, json_schema_extra={"vmanage_key": "svc-type"})
    interface_pair: List[InterfacePair] = Field(
        json_schema_extra={"data_path": ["ha-pairs"], "vmanage_key": "interface-pair"}
    )
    auth_required: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "auth-required"})
    xff_forward_enabled: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "xff-forward-enabled"})
    ofw_enabled: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "ofw-enabled"})
    ips_control: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "ips-control"})
    caution_enabled: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "caution-enabled"})
    primary_data_center: Optional[str] = Field("Auto", json_schema_extra={"vmanage_key": "primary-data-center"})
    secondary_data_center: Optional[str] = Field("Auto", json_schema_extra={"vmanage_key": "secondary-data-center"})
    ip: Optional[bool] = None
    idle_time: Optional[int] = Field(DEFAULT_SERVICE_IDLE_TIME, json_schema_extra={"vmanage_key": "idle-time"})
    display_time_unit: Optional[DisplayTimeUnit] = Field(
        DisplayTimeUnit.MINUTE, json_schema_extra={"vmanage_key": "display-time-unit"}
    )
    ip_enforced_for_known_browsers: Optional[bool] = Field(
        False, json_schema_extra={"vmanage_key": "ip-enforced-for-known-browsers"}
    )
    refresh_time: Optional[int] = Field(DEFAULT_SERVICE_REFRESH_TIME, json_schema_extra={"vmanage_key": "refresh-time"})
    refresh_time_unit: Optional[RefreshTimeUnit] = Field(
        RefreshTimeUnit.MINUTE, json_schema_extra={"vmanage_key": "refresh-time-unit"}
    )
    enabled: Optional[bool] = None
    block_internet_until_accepted: Optional[bool] = Field(
        False, json_schema_extra={"vmanage_key": "block-internet-until-accepted"}
    )
    force_ssl_inspection: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "force-ssl-inspection"})
    timeout: Optional[int] = None
    data_center_primary: Optional[str] = Field("Auto", json_schema_extra={"vmanage_key": "data-center-primary"})
    data_center_secondary: Optional[str] = Field("Auto", json_schema_extra={"vmanage_key": "data-center-secondary"})
    model_config = ConfigDict(populate_by_name=True)


class TrackerType(str, Enum):
    SIG = "SIG"


class Tracker(FeatureTemplateValidator):
    name: str
    endpoint_api_url: str = Field(json_schema_extra={"vmanage_key": "endpoint-api-url"})
    threshold: Optional[int] = DEFAULT_TRACKER_THRESHOLD
    interval: Optional[int] = DEFAULT_TRACKER_INTERVAL
    multiplier: Optional[int] = DEFAULT_TRACKER_MULTIPLIER
    tracker_type: TrackerType = Field(json_schema_extra={"vmanage_key": "tracker-type"})
    model_config = ConfigDict(populate_by_name=True)


class CiscoSecureInternetGatewayModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    vpn_id: int = Field(DEFAULT_SIG_VPN_ID, json_schema_extra={"vmanage_key": "vpn-id"})
    interface: List[Interface]
    service: List[Service]
    tracker_src_ip: Optional[ipaddress.IPv4Interface] = Field(
        default=None, json_schema_extra={"vmanage_key": "tracker-src-ip"}
    )
    tracker: Optional[List[Tracker]] = None

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_secure_internet_gateway"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from pathlib import Path
from typing import ClassVar, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.feature_template import FeatureTemplate


class Protocol(str, Enum):
    DTLS: str = "dtls"
    TLS: str = "tls"


class SecurityvSmart(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Optional[Protocol] = Field(default=None, json_schema_extra={"data_path": ["control"]})
    tls_port: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "tls-port", "data_path": ["control"]}
    )
    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "security-vsmart"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from unittest.mock import patch

from catalystwan.api.task_status_api import Task
from catalystwan.endpoints.configuration_dashboard_status import ConfigurationDashboardStatus, TaskData
from catalystwan.exceptions import TaskValidationError


class TestTaskStatusApi(unittest.TestCase):
    @patch("catalystwan.session.ManagerSession")
    def setUp(self, mock_session):
        self.task = Task(mock_session, "task_id")
        self.success_response = {
            "data": [
                {
                    "local-system-ip": "local_ip",
                    "statusType": "reboot",
                    "activity": [],
                    "system-ip": "system_ip",
                    "site-id": "siteid",
                    "uuid": "dev-uuid",
                    "@rid": 1211,
                    "personality": "vedge",
                    "processId": "processid",
                    "actionConfig": "",
                    "device-type": "vedge",
                    "action": "reboot",
                    "startTime": 1685440088317,
                    "reachability": "reachable",
                    "order": 0,
                    "vmanageIP": "vmanage_ip",
                    "host-name": "vm1",
                    "version": "vmanage-version",
                    "deviceID": "deviceid",
                    "statusId": "success",
                    "currentActivity": "Done - Reboot",
                    "deviceModel": "vedge-cloud",
                    "validity": "valid",
                    "requestStatus": "received",
                    "status": "Success",
                }
            ],
            "validation": {
                "statusType": "reboot",
                "activity": [],
                "vmanageIP": "vmanage-ip",
                "system-ip": "Validation",
                "deviceID": "Validation",
                "uuid": "Validation",
                "@rid": 747,
                "statusId": "validation_success",
                "processId": "reboot-9fc30834-cc46-47c5-83c4-0b837cf84f1a",
                "actionConfig": "{}",
                "currentActivity": "Done - Validation",
                "action": "reboot",
                "startTime": 1685440057748,
                "requestStatus": "received",
                "status": "Validation success",
                "order": 0,
            },
            "summary": {
                "action": "reboot",
                "name": "Reboot",
                "detailsURL": "/dataservice/device/action/status",
                "startTime": "1685440057829",
                "endTime": "1685440179295",
                "userSessionUserName": "admin",
                "userSessionIP": "10.0.1.1",
                "tenantName": "DefaultTenant",
                "total": 1,
                "status": "done",
                "count": {"Success": 1},
            },
            "isCancelEnabled": True,
            "isParallelExecutionEnabled": True,
        }
        self.empty_data = {
            "data": [],
            "validation": {
                "statusType": "reboot",
                "activity": [],
                "vmanageIP": "vmanage-ip",
                "system-ip": "Validation",
                "deviceID": "Validation",
                "uuid": "Validation",
                "@rid": 747,
                "statusId": "validation_success",
                "processId": "reboot-9fc30834-cc46-47c5-83c4-0b837cf84f1a",
                "actionConfig": "{}",
                "currentActivity": "Done - Validation",
                "action": "reboot",
                "startTime": 1685440057748,
                "requestStatus": "received",
                "status": "Validation success",
                "order": 0,
            },
            "summary": {
                "action": "reboot",
                "name": "Reboot",
                "detailsURL": "/dataservice/device/action/status",
                "startTime": "1685440057829",
                "endTime": "1685440179295",
                "userSessionUserName": "admin",
                "userSessionIP": "user-session-ip",
                "tenantName": "DefaultTenant",
                "total": 1,
                "status": "done",
                "count": {"Success": 1},
            },
            "isCancelEnabled": True,
            "isParallelExecutionEnabled": True,
        }
        self.no_data = {
            "validation": {
                "statusType": "reboot",
                "activity": [],
                "vmanageIP": "ip",
                "system-ip": "Validation",
                "deviceID": "Validation",
                "uuid": "Validation",
                "@rid": 747,
                "statusId": "validation_success",
                "processId": "reboot-9fc30834-cc46-47c5-83c4-0b837cf84f1a",
                "actionConfig": "{}",
                "currentActivity": "Done - Validation",
                "action": "reboot",
                "startTime": 1685440057748,
                "requestStatus": "received",
                "status": "Validation success",
                "order": 0,
            },
            "summary": {
                "action": "reboot",
                "name": "Reboot",
                "detailsURL": "/dataservice/device/action/status",
                "startTime": "1685440057829",
                "endTime": "1685440179295",
                "userSessionUserName": "admin",
                "userSessionIP": "ip",
                "tenantName": "DefaultTenant",
                "total": 1,
                "status": "done",
                "count": {"Success": 1},
            },
            "isCancelEnabled": True,
            "isParallelExecutionEnabled": True,
        }
        self.no_validation = {
            "data": [
                {
                    "local-system-ip": "local_ip",
                    "statusType": "reboot",
                    "activity": [],
                    "system-ip": "system_ip",
                    "site-id": "siteid",
                    "uuid": "dev-uuid",
                    "@rid": 1211,
                    "personality": "vedge",
                    "processId": "processid",
                    "actionConfig": "",
                    "device-type": "vedge",
                    "action": "reboot",
                    "startTime": 1685440088317,
                    "reachability": "reachable",
                    "order": 0,
                    "vmanageIP": "vmanage_ip",
                    "host-name": "vm1",
                    "version": "vmanage-version",
                    "deviceID": "deviceid",
                    "statusId": "success",
                    "currentActivity": "Done - Reboot",
                    "deviceModel": "vedge-cloud",
                    "validity": "valid",
                    "requestStatus": "received",
                    "status": "Success",
                }
            ],
            "summary": {
                "action": "reboot",
                "name": "Reboot",
                "detailsURL": "/dataservice/device/action/status",
                "startTime": "1685440057829",
                "endTime": "1685440179295",
                "userSessionUserName": "admin",
                "userSessionIP": "10.0.1.1",
                "tenantName": "DefaultTenant",
                "total": 1,
                "status": "done",
                "count": {"Success": 1},
            },
            "isCancelEnabled": True,
            "isParallelExecutionEnabled": True,
        }
        self.validation_failure = {
            "validation": {
                "statusType": "reboot",
                "activity": [],
                "vmanageIP": "vmanage-ip",
                "system-ip": "Validation",
                "deviceID": "Validation",
                "uuid": "Validation",
                "@rid": 747,
                "statusId": "validation_success",
                "processId": "reboot-9fc30834-cc46-47c5-83c4-0b837cf84f1a",
                "actionConfig": "{}",
                "currentActivity": "Done - Validation",
                "action": "reboot",
                "startTime": 1685440057748,
                "requestStatus": "received",
                "status": "Validation failure",
                "order": 0,
            },
            "summary": {
                "action": "reboot",
                "name": "Reboot",
                "detailsURL": "/dataservice/device/action/status",
                "startTime": "1685440057829",
                "endTime": "1685440179295",
                "userSessionUserName": "admin",
                "userSessionIP": "10.0.1.1",
                "tenantName": "DefaultTenant",
                "total": 1,
                "status": "done",
                "count": {"Success": 1},
            },
            "isCancelEnabled": True,
            "isParallelExecutionEnabled": True,
        }
        self.response_with_action_config_as_dict = {
            "data": [
                {
                    "local-system-ip": "local_ip",
                    "statusType": "reboot",
                    "activity": [],
                    "system-ip": "system_ip",
                    "site-id": "siteid",
                    "uuid": "dev-uuid",
                    "@rid": 1211,
                    "personality": "vedge",
                    "processId": "processid",
                    "actionConfig": {
                        "devices": {
                            "deviceIP": "",
                            "deviceId": "",
                            "version": "",
                            "order": "",
                            "isNutellaMigration": False,
                        }
                    },
                    "device-type": "vedge",
                    "action": "reboot",
                    "startTime": 1685440088317,
                    "reachability": "reachable",
                    "order": 0,
                    "vmanageIP": "vmanage_ip",
                    "host-name": "vm1",
                    "version": "vmanage-version",
                    "deviceID": "deviceid",
                    "statusId": "success",
                    "currentActivity": "Done - Reboot",
                    "deviceModel": "vedge-cloud",
                    "validity": "valid",
                    "requestStatus": "received",
                    "status": "Success",
                }
            ],
            "validation": {
                "statusType": "reboot",
                "activity": [],
                "vmanageIP": "vmanage-ip",
                "system-ip": "Validation",
                "deviceID": "Validation",
                "uuid": "Validation",
                "@rid": 747,
                "statusId": "validation_success",
                "processId": "reboot-9fc30834-cc46-47c5-83c4-0b837cf84f1a",
                "actionConfig": {
                    "devices": {"deviceIP": "", "deviceId": "", "version": "", "order": "", "isNutellaMigration": False}
                },
                "currentActivity": "Done - Validation",
                "action": "reboot",
                "startTime": 1685440057748,
                "requestStatus": "received",
                "status": "Validation success",
                "order": 0,
            },
            "summary": {
                "action": "reboot",
                "name": "Reboot",
                "detailsURL": "/dataservice/device/action/status",
                "startTime": "1685440057829",
                "endTime": "1685440179295",
                "userSessionUserName": "admin",
                "userSessionIP": "10.0.1.1",
                "tenantName": "DefaultTenant",
                "total": 1,
                "status": "done",
                "count": {"Success": 1},
            },
            "isCancelEnabled": True,
            "isParallelExecutionEnabled": True,
        }

    @patch.object(Task, "_Task__check_validation_status")
    @patch.object(ConfigurationDashboardStatus, "find_status")
    def test_wait_for_completed_success(self, mock_task_response, mock_validation):
        # Arrange
        mock_task_response.return_value = TaskData.model_validate(self.success_response)

        # Act
        answer = self.task.wait_for_completed(interval_seconds=1).result

        # Assert
        self.assertEqual(answer, True)

    @patch.object(Task, "_Task__check_validation_status")
    @patch.object(ConfigurationDashboardStatus, "find_status")
    def test_wait_for_completed_empty_data(self, mock_task_response, mock_validation):
        # Data is empty, and then response is success

        # Arrange
        mock_task_response.side_effect = [
            TaskData.model_validate(self.empty_data),
            TaskData.model_validate(self.success_response),
        ]

        # Act
        answer = self.task.wait_for_completed(timeout_seconds=2, interval_seconds=1).result

        # Assert
        self.assertEqual(answer, True)

    @patch.object(Task, "_Task__check_validation_status")
    @patch.object(ConfigurationDashboardStatus, "find_status")
    def test_wait_for_completed_with_action_config_as_dict(self, mock_task_response, mock_validation):
        # Arrange
        mock_task_response.return_value = TaskData.model_validate(self.response_with_action_config_as_dict)

        # Act
        answer = self.task.wait_for_completed(interval_seconds=1).result

        # Assert
        self.assertEqual(answer, True)

    @patch.object(Task, "_Task__check_validation_status")
    @patch.object(ConfigurationDashboardStatus, "find_status")
    def test_wait_for_completed_no_data(self, mock_task_response, mock_validation):
        # No data in first call, and then response is success

        # Arrange
        mock_task_response.side_effect = [
            TaskData.model_validate(self.no_data),
            TaskData.model_validate(self.success_response),
        ]

        # Act
        answer = self.task.wait_for_completed(timeout_seconds=2, interval_seconds=1).result

        # Assert
        self.assertEqual(answer, True)

    @patch.object(ConfigurationDashboardStatus, "find_status")
    def test_wait_for_completed_no_validation_field(self, mock_task_response_response):
        # Arrange
        mock_task_response_response.return_value = TaskData.model_validate(self.no_validation)

        # Act
        answer = self.task.wait_for_completed(interval_seconds=1).result

        # Assert
        self.assertEqual(answer, True)

    @patch.object(ConfigurationDashboardStatus, "find_status")
    def test_wait_for_completed_raise_TaskValidationError(self, mock_task_response_response):
        # Arrange
        mock_task_response_response.return_value = TaskData.model_validate(self.validation_failure)

        # Act&Assert
        self.assertRaises(TaskValidationError, self.task.wait_for_completed)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Optional

from pydantic import AliasPath, BaseModel, Field, field_validator, model_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global
from catalystwan.models.common import InterfaceType, check_fields_exclusive


class SecurityZoneListEntry(BaseModel):
    vpn: Optional[Global[str]] = Field(default=None, description="0-65530 single number")
    interface: Optional[Global[InterfaceType]] = None

    @field_validator("vpn")
    @classmethod
    def check_vpn_range(cls, vpn: Global[str]):
        assert 0 <= int(vpn.value) <= 65530
        return vpn

    @model_validator(mode="after")
    def check_vpn_xor_interface(self):
        check_fields_exclusive(self.__dict__, {"vpn", "interface"}, True)
        return self


class SecurityZoneListParcel(_ParcelBase):
    entries: List[SecurityZoneListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_interface(self, interface: InterfaceType):
        self.entries.append(
            SecurityZoneListEntry(
                interface=as_global(interface, InterfaceType),
            )
        )

    def add_vpn(self, vpn: str):
        self.entries.append(
            SecurityZoneListEntry(
                vpn=as_global(vpn, InterfaceType),
            )
        )



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from datetime import datetime
from typing import Optional

from pydantic import AliasChoices, BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, get
from catalystwan.typed_list import DataSequence


class RebootEntry(BaseModel):
    model_config = ConfigDict(extra="allow")
    reason: Optional[str] = Field(default=None, validation_alias=AliasChoices("reboot_reason", "reload-desc"))
    severity: Optional[str] = Field(default=None, validation_alias="reload-severity", description=">=17.13 only")
    category: Optional[str] = Field(default=None, validation_alias="reload-category", description=">=17.13 only")
    date_time: datetime = Field(
        validation_alias=AliasChoices("reboot_date_time-date", "reboot_date_time", "reload-time-date", "reload-time")
    )
    last_updated: datetime = Field(validation_alias="lastupdated")
    vdevice_name: str = Field(validation_alias="vdevice-name")
    vdevice_host_name: str = Field(validation_alias="vdevice-host-name")


class RealTimeMonitoringRebootHistory(APIEndpoints):
    @get("/device/reboothistory", "data")
    def create_reboot_history_list(self, params: dict) -> DataSequence[RebootEntry]:
        ...

    def create_synced_reboot_history_list(self):
        # GET /device/reboothistory/synced
        ...

    def get_reboot_history_details(self):
        # GET /device/reboothistory/details
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from pathlib import PurePath
from typing import TYPE_CHECKING, Dict, List, Union

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints.configuration.software_actions import SoftwareImageDetails
from catalystwan.endpoints.configuration_device_actions import (
    InstallLxcImage,
    LxcActivateDevice,
    LxcUpgradeDevice,
    PartitionDevice,
)
from catalystwan.endpoints.configuration_device_inventory import DeviceDetailsResponse
from catalystwan.exceptions import ImageNotInRepositoryError
from catalystwan.typed_list import DataSequence
from catalystwan.utils.upgrades_helper import SoftwarePackageUploadPayload

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

logger = logging.getLogger(__name__)


class DeviceSoftwareRepository(BaseModel):
    model_config = ConfigDict(extra="ignore")

    installed_versions: List[str] = Field(default_factory=list)
    available_versions: List[str] = Field(
        default_factory=list, serialization_alias="availableVersions", validation_alias="availableVersions"
    )
    current_version: str = Field(
        default="",
        serialization_alias="version",
        validation_alias="version",
        description="Current active version of software on device",
    )
    default_version: str = Field(default="", serialization_alias="defaultVersion", validation_alias="defaultVersion")
    device_id: str = Field(default="", serialization_alias="uuid", validation_alias="uuid")


class RepositoryAPI:
    """
    API methods to get information about images and devices software versions

    Usage example:
        # Create session
        session = create_manager_session(...)

        # Upload image
        software_image = <path_to_your_image>
        RepositoryAPI(provider).upload_image(software_image)
    """

    def __init__(
        self,
        session: ManagerSession,
    ):
        self.session = session

    def get_all_software_images(self) -> DataSequence[SoftwareImageDetails]:
        """
        Get all info about all software images stored in Vmanage repository

        Returns:
            list: software images list
        """
        software_images = self.session.endpoints.configuration_software_actions.get_list_of_all_images()
        return software_images

    def get_all_virtual_images(self) -> DataSequence[SoftwareImageDetails]:
        """
        Get all info about all software images stored in Vmanage repository

        Returns:
            list: software images list
        """
        software_images = self.session.endpoints.configuration_software_actions.get_list_of_all_images(
            "imageType=virtualmachine"
        )
        return software_images

    def get_devices_versions_repository(self) -> Dict[str, DeviceSoftwareRepository]:
        """
        Create DeviceSoftwareRepository dataclass,
        which cointains information about all possible version types for certain devices

        Returns:
            Dict[str, DeviceSoftwareRepository]: Dictionary containing all versions
            information
        """

        controllers_versions_info = self.session.endpoints.configuration_device_actions.get_list_of_installed_devices(
            device_type="controller"
        )
        edges_versions_info = self.session.endpoints.configuration_device_actions.get_list_of_installed_devices(
            device_type="vedge"
        )
        vmanages_versions_info = self.session.endpoints.configuration_device_actions.get_list_of_installed_devices(
            device_type="vmanage"
        )
        devices_versions_repository = {}
        for device in controllers_versions_info + edges_versions_info + vmanages_versions_info:
            device_software_repository = DeviceSoftwareRepository(**device.model_dump(by_alias=True))
            device_software_repository.installed_versions = [a for a in device_software_repository.available_versions]
            device_software_repository.installed_versions.append(device_software_repository.current_version)
            devices_versions_repository[device_software_repository.device_id] = device_software_repository
        return devices_versions_repository

    def get_image_version(self, software_image: str) -> Union[str, None]:
        """
        Get proper software image version, based on name in available files.

        If software_image detected in available files, but doesn't include version_name, software_image won't be used.

        Args:
            software_image (str): path to software image

        Returns:
            Union[str, None]: image version or None
        """

        image_name = PurePath(software_image).name
        software_images = self.get_all_software_images()
        for image in software_images:
            if image.available_files and image_name in image.available_files:
                if image.version_name and not image.version_name == "--":
                    return image.version_name
                logger.warning(
                    f"Detected image {image_name} in available files has version_name: {image.version_name} as value."
                    "Image will not be used and image version won't be returned."
                )
        logger.error(f"Software image {image_name} is not in available images")
        return None

    def get_remote_image(
        self, remote_image_filename: str, remote_server_name: str
    ) -> Union[SoftwareImageDetails, None]:
        """
        Get remote software image details, based on name in available files and remote server name.

        Args:
            remote_image_filename (str): path to software image on remote server
            remote_server_name (str): remote server name

        Returns:
            Union[SoftwareImageDetails, None]: remote image image details
        """

        image_name = PurePath(remote_image_filename).name
        software_images = self.get_all_software_images()
        for image_details in software_images:
            if (
                image_details.available_files
                and image_details.version_type
                and image_name in image_details.available_files
                and remote_server_name in image_details.version_type
            ):
                if not (image_details.remote_server_id and image_details.version_id):
                    raise ValueError(
                        f"Requested image: '{image_name}' does not include include required fields for this operation:"
                        f"image_details.remote_server_id - (current value: {image_details.remote_server_id})"
                        f"image_details.version_id - (current value: {image_details.version_id})"
                    )
                return image_details
        logger.error(
            f"Software image {image_name} is not in available in images from remote server {remote_server_name}"
        )
        return None

    def upload_image(self, image_path: str) -> None:
        """
        Upload software image ('tar.gz' or 'SPA.bin') to vManage software repository

        Args:
            image_path (str): path to software image

        Returns:
            str: Response status code
        """
        self.session.endpoints.configuration_device_software_update.upload_software_to_manager(
            payload=SoftwarePackageUploadPayload(image_path=image_path)
        )

    def delete_image(self, image_name: str) -> None:
        """
        Delete image from vManage software repository

        Args:
            image_name (str): image name (in available files)

        Raises:
            ImageNotInRepositoryError: raise error if image not in repository

        Returns:
            int: Reponse status code
        """
        for image in self.get_all_software_images():
            if image.available_files and image_name in image.available_files:
                version_id = image.version_id
                self.session.endpoints.configuration_software_actions.delete_software_from_software_repository(
                    version_id=version_id
                )
                return None
        raise ImageNotInRepositoryError(f"Image: {image_name} is not the vManage software repository")


class DeviceVersions:
    """
    Methods to prepare devices list for payload
    """

    def __init__(self, session: ManagerSession):
        self.repository = RepositoryAPI(session)

    def _validate_devices_required_fields(self, devices: DataSequence[DeviceDetailsResponse]):
        for device in devices:
            if not device.uuid or not device.local_system_ip:
                raise ValueError(
                    f"Provided device '{device.host_name}' doesn't include required fields for this operation:"
                    f"device.uuid (current value: {device.uuid})"
                    f"device.device_ip (current value: {device.device_ip})"
                )

    def get_lxcactivate_device_list(
        self,
        version_to_set_up: str,
        devices: DataSequence[DeviceDetailsResponse],
    ) -> DataSequence[LxcActivateDevice]:
        self._validate_devices_required_fields(devices)
        install_image_payload = [
            InstallLxcImage(
                network_function_type="app-hosting",
                version_name=version_to_set_up,
                version_type_name="UTD-Snort-Feature",
            )
        ]
        devices_payload = DataSequence(
            LxcActivateDevice,
            [
                LxcActivateDevice(
                    device_id=str(device.uuid),
                    device_ip=str(device.local_system_ip),
                    install_images=install_image_payload,
                )
                for device in devices
            ],  # type: ignore
        )

        return devices_payload

    def get_lxcupgrade_device_list(
        self,
        version_to_set_up: str,
        devices: DataSequence[DeviceDetailsResponse],
    ) -> DataSequence[LxcUpgradeDevice]:
        self._validate_devices_required_fields(devices)
        install_image_payload = [InstallLxcImage(version_name=version_to_set_up)]
        devices_payload = DataSequence(
            LxcUpgradeDevice,
            [
                LxcUpgradeDevice(
                    device_id=str(device.uuid),
                    device_ip=str(device.local_system_ip),
                    install_images=install_image_payload,
                )
                for device in devices
            ],  # type: ignore
        )

        return devices_payload

    def _get_device_list_in(
        self, version_to_set_up: str, devices: DataSequence[DeviceDetailsResponse], version_type: str
    ) -> DataSequence[PartitionDevice]:
        """
        Create devices payload list included requested version, if requested version
        is in specified version type

        Args:
            version_to_set_up (str): requested version
            devices List[DeviceDetailsResponse]: list of Device dataclass instances
            version_type: type of version (installed, available, etc.)

        Returns:
            list : list of devices
        """
        self._validate_devices_required_fields(devices)
        devices_payload = DataSequence(
            PartitionDevice,
            [PartitionDevice(device_id=str(device.uuid), device_ip=str(device.local_system_ip)) for device in devices],
            # type: ignore
        )

        all_dev_versions = self.repository.get_devices_versions_repository()
        for device in devices_payload:
            device_versions = getattr(all_dev_versions[device.device_id], version_type)
            try:
                for version in device_versions:
                    if version_to_set_up in version:
                        device.version = version
                        break
            except IndexError:
                logger.error(
                    f"Software version {version_to_set_up} for {device} is not included in {version_type}."
                    "Action for that device is not going to proceed."
                )
        return devices_payload

    def get_device_list_in_installed(
        self, version_to_set_up: str, devices: DataSequence[DeviceDetailsResponse]
    ) -> DataSequence[PartitionDevice]:
        """
        Create devices payload list included requested version, if requested version
        is in installed versions

        Args:
            version_to_set_up (str): requested version
            devices (List[DeviceDetailsResponse]): devices on which action going to be performed

        Returns:
            list : list of devices
        """
        return self._get_device_list_in(version_to_set_up, devices, "installed_versions")

    def get_device_available(
        self, version_to_set_up: str, devices: DataSequence[DeviceDetailsResponse]
    ) -> DataSequence[PartitionDevice]:
        """
        Create devices payload list included requested, if requested version
        is in available versions

        Args:
            version_to_set_up (str): requested version
            devices (List[DeviceDetailsResponse]): devices on which action going to be performed


        Returns:
            list : list of devices
        """
        return self._get_device_list_in(version_to_set_up, devices, "available_versions")

    def _get_devices_chosen_version(
        self, devices: DataSequence[DeviceDetailsResponse], version_type: str
    ) -> DataSequence[PartitionDevice]:
        """
        Create devices payload list included software version key
        for every device in devices list

        Args:
            version_to_set_up (str): requested version
            devices (List[DeviceDetailsResponse]): devices on which action going to be performed

        Returns:
            list : list of devices
        """
        self._validate_devices_required_fields(devices)

        devices_payload = DataSequence(
            PartitionDevice,
            [
                PartitionDevice(device_id=device.uuid, device_ip=device.local_system_ip)  # type: ignore
                for device in devices
            ],  # type: ignore
        )
        all_dev_versions = self.repository.get_devices_versions_repository()
        for device in devices_payload:
            device.version = getattr(all_dev_versions[device.device_id], version_type)
        return devices_payload

    def get_devices_current_version(
        self, devices: DataSequence[DeviceDetailsResponse]
    ) -> DataSequence[PartitionDevice]:
        """
        Create devices payload list included current software version key
        for every device in devices list

        Args:
            version_to_set_up (str): requested version
            devices (List[DeviceDetailsResponse]): devices on which action going to be performed

        Returns:
            list : list of devices
        """

        return self._get_devices_chosen_version(devices, "current_version")

    def get_devices_available_versions(
        self, devices: DataSequence[DeviceDetailsResponse]
    ) -> DataSequence[PartitionDevice]:
        """
        Create devices payload list included available software versions key
        for every device in devices list

        Args:
            devices (List[DeviceDetailsResponse]): devices on which action going to be performed

        Returns:
            list : list of devices
        """

        return self._get_devices_chosen_version(devices, "available_versions")

    def get_device_list(self, devices: DataSequence[DeviceDetailsResponse]) -> List[PartitionDevice]:
        self._validate_devices_required_fields(devices)

        return [
            PartitionDevice(device_id=device.uuid, device_ip=device.device_ip) for device in devices  # type: ignore
        ]



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from typing import TYPE_CHECKING

from catalystwan.endpoints.administration_user_and_group import AdministrationUserAndGroup
from catalystwan.endpoints.certificate_management_device import CertificateManagementDevice
from catalystwan.endpoints.certificate_management_vmanage import CertificateManagementVManage
from catalystwan.endpoints.client import Client
from catalystwan.endpoints.cluster_management import ClusterManagement
from catalystwan.endpoints.configuration.device.software_update import ConfigurationDeviceSoftwareUpdate
from catalystwan.endpoints.configuration.disaster_recovery import ConfigurationDisasterRecovery
from catalystwan.endpoints.configuration.feature_profile.sdwan.system import SystemFeatureProfile
from catalystwan.endpoints.configuration.feature_profile.sdwan.transport import TransportFeatureProfile
from catalystwan.endpoints.configuration.policy.definition.access_control_list import ConfigurationPolicyAclDefinition
from catalystwan.endpoints.configuration.policy.definition.access_control_list_ipv6 import (
    ConfigurationPolicyAclIPv6Definition,
)
from catalystwan.endpoints.configuration.policy.definition.control import ConfigurationPolicyControlDefinition
from catalystwan.endpoints.configuration.policy.definition.device_access import (
    ConfigurationPolicyDeviceAccessDefinition,
)
from catalystwan.endpoints.configuration.policy.definition.device_access_ipv6 import (
    ConfigurationPolicyDeviceAccessIPv6Definition,
)
from catalystwan.endpoints.configuration.policy.definition.hub_and_spoke import ConfigurationPolicyHubAndSpokeDefinition
from catalystwan.endpoints.configuration.policy.definition.mesh import ConfigurationPolicyMeshDefinition
from catalystwan.endpoints.configuration.policy.definition.qos_map import ConfigurationPolicyQoSMapDefinition
from catalystwan.endpoints.configuration.policy.definition.rewrite import ConfigurationPolicyRewriteRuleDefinition
from catalystwan.endpoints.configuration.policy.definition.rule_set import ConfigurationPolicyRuleSetDefinition
from catalystwan.endpoints.configuration.policy.definition.security_group import (
    ConfigurationPolicySecurityGroupDefinition,
)
from catalystwan.endpoints.configuration.policy.definition.traffic_data import ConfigurationPolicyDataDefinition
from catalystwan.endpoints.configuration.policy.definition.vpn_membership import (
    ConfigurationPolicyVPNMembershipGroupDefinition,
)
from catalystwan.endpoints.configuration.policy.definition.zone_based_firewall import (
    ConfigurationPolicyZoneBasedFirewallDefinition,
)
from catalystwan.endpoints.configuration.policy.list.app import ConfigurationPolicyApplicationList
from catalystwan.endpoints.configuration.policy.list.app_probe import ConfigurationPolicyAppProbeClassList
from catalystwan.endpoints.configuration.policy.list.as_path import ConfigurationPolicyASPathList
from catalystwan.endpoints.configuration.policy.list.class_map import ConfigurationPolicyForwardingClassList
from catalystwan.endpoints.configuration.policy.list.color import ConfigurationPolicyColorList
from catalystwan.endpoints.configuration.policy.list.community import ConfigurationPolicyCommunityList
from catalystwan.endpoints.configuration.policy.list.data_ipv6_prefix import ConfigurationPolicyDataIPv6PrefixList
from catalystwan.endpoints.configuration.policy.list.data_prefix import ConfigurationPolicyDataPrefixList
from catalystwan.endpoints.configuration.policy.list.expanded_community import ConfigurationPolicyExpandedCommunityList
from catalystwan.endpoints.configuration.policy.list.fqdn import ConfigurationPolicyFQDNList
from catalystwan.endpoints.configuration.policy.list.geo_location import ConfigurationPolicyGeoLocationList
from catalystwan.endpoints.configuration.policy.list.ips_signature import ConfigurationPolicyIPSSignatureList
from catalystwan.endpoints.configuration.policy.list.ipv6_prefix import ConfigurationPolicyIPv6PrefixList
from catalystwan.endpoints.configuration.policy.list.local_app import ConfigurationPolicyLocalAppList
from catalystwan.endpoints.configuration.policy.list.local_domain import ConfigurationPolicyLocalDomainList
from catalystwan.endpoints.configuration.policy.list.mirror import ConfigurationPolicyMirrorList
from catalystwan.endpoints.configuration.policy.list.policer import ConfigurationPolicyPolicerClassList
from catalystwan.endpoints.configuration.policy.list.port import ConfigurationPolicyPortList
from catalystwan.endpoints.configuration.policy.list.preferred_color_group import ConfigurationPreferredColorGroupList
from catalystwan.endpoints.configuration.policy.list.prefix import ConfigurationPolicyPrefixList
from catalystwan.endpoints.configuration.policy.list.protocol_name import ConfigurationPolicyProtocolNameList
from catalystwan.endpoints.configuration.policy.list.region import ConfigurationPolicyRegionList
from catalystwan.endpoints.configuration.policy.list.site import ConfigurationPolicySiteList
from catalystwan.endpoints.configuration.policy.list.sla import ConfigurationPolicySLAClassList
from catalystwan.endpoints.configuration.policy.list.tloc import ConfigurationPolicyTLOCList
from catalystwan.endpoints.configuration.policy.list.url_allow_list import ConfigurationPolicyURLAllowList
from catalystwan.endpoints.configuration.policy.list.url_block_list import ConfigurationPolicyURLBlockList
from catalystwan.endpoints.configuration.policy.list.vpn import ConfigurationPolicyVPNList
from catalystwan.endpoints.configuration.policy.list.zone import ConfigurationPolicyZoneList
from catalystwan.endpoints.configuration.policy.security_template import ConfigurationSecurityTemplatePolicy
from catalystwan.endpoints.configuration.policy.vedge_template import ConfigurationVEdgeTemplatePolicy
from catalystwan.endpoints.configuration.policy.vsmart_template import ConfigurationVSmartTemplatePolicy
from catalystwan.endpoints.configuration.software_actions import ConfigurationSoftwareActions
from catalystwan.endpoints.configuration_dashboard_status import ConfigurationDashboardStatus
from catalystwan.endpoints.configuration_device_actions import ConfigurationDeviceActions
from catalystwan.endpoints.configuration_device_inventory import ConfigurationDeviceInventory
from catalystwan.endpoints.configuration_device_template import ConfigurationDeviceTemplate
from catalystwan.endpoints.configuration_feature_profile import (
    ConfigurationFeatureProfile,
    SDRoutingConfigurationFeatureProfile,
)
from catalystwan.endpoints.configuration_group import ConfigurationGroup
from catalystwan.endpoints.configuration_settings import ConfigurationSettings
from catalystwan.endpoints.misc import MiscellaneousEndpoints
from catalystwan.endpoints.monitoring.device_details import MonitoringDeviceDetails
from catalystwan.endpoints.monitoring.security_policy import MonitoringSecurityPolicy
from catalystwan.endpoints.monitoring.server_info import ServerInfo
from catalystwan.endpoints.monitoring.status import MonitoringStatus
from catalystwan.endpoints.real_time_monitoring.reboot_history import RealTimeMonitoringRebootHistory
from catalystwan.endpoints.sdavc_cloud_connector import SDAVCCloudConnector
from catalystwan.endpoints.tenant_backup_restore import TenantBackupRestore
from catalystwan.endpoints.tenant_management import TenantManagement
from catalystwan.endpoints.tenant_migration import TenantMigration
from catalystwan.endpoints.troubleshooting_tools.device_connectivity import TroubleshootingToolsDeviceConnectivity
from catalystwan.endpoints.url_monitoring import UrlMonitoring

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class ConfigurationPolicyListContainer:
    def __init__(self, session: ManagerSession):
        self.app = ConfigurationPolicyApplicationList(session)
        self.app_probe = ConfigurationPolicyAppProbeClassList(session)
        self.as_path = ConfigurationPolicyASPathList(session)
        self.class_map = ConfigurationPolicyForwardingClassList(session)
        self.color = ConfigurationPolicyColorList(session)
        self.community = ConfigurationPolicyCommunityList(session)
        self.data_ipv6_prefix = ConfigurationPolicyDataIPv6PrefixList(session)
        self.data_prefix = ConfigurationPolicyDataPrefixList(session)
        self.expanded_community = ConfigurationPolicyExpandedCommunityList(session)
        self.fqdn = ConfigurationPolicyFQDNList(session)
        self.geo_location = ConfigurationPolicyGeoLocationList(session)
        self.ips_signature = ConfigurationPolicyIPSSignatureList(session)
        self.ipv6_prefix = ConfigurationPolicyIPv6PrefixList(session)
        self.local_app = ConfigurationPolicyLocalAppList(session)
        self.local_domain = ConfigurationPolicyLocalDomainList(session)
        self.mirror = ConfigurationPolicyMirrorList(session)
        self.policer = ConfigurationPolicyPolicerClassList(session)
        self.port = ConfigurationPolicyPortList(session)
        self.preferred_color_group = ConfigurationPreferredColorGroupList(session)
        self.prefix = ConfigurationPolicyPrefixList(session)
        self.protocol_name = ConfigurationPolicyProtocolNameList(session)
        self.region = ConfigurationPolicyRegionList(session)
        self.site = ConfigurationPolicySiteList(session)
        self.sla = ConfigurationPolicySLAClassList(session)
        self.tloc = ConfigurationPolicyTLOCList(session)
        self.url_block_list = ConfigurationPolicyURLBlockList(session)
        self.url_allow_list = ConfigurationPolicyURLAllowList(session)
        self.vpn = ConfigurationPolicyVPNList(session)
        self.zone = ConfigurationPolicyZoneList(session)


class ConfigurationPolicyDefinitionContainer:
    def __init__(self, session: ManagerSession):
        self.data = ConfigurationPolicyDataDefinition(session)
        self.rule_set = ConfigurationPolicyRuleSetDefinition(session)
        self.security_group = ConfigurationPolicySecurityGroupDefinition(session)
        self.zone_based_firewall = ConfigurationPolicyZoneBasedFirewallDefinition(session)
        self.qos_map = ConfigurationPolicyQoSMapDefinition(session)
        self.rewrite = ConfigurationPolicyRewriteRuleDefinition(session)
        self.control = ConfigurationPolicyControlDefinition(session)
        self.vpn_membership = ConfigurationPolicyVPNMembershipGroupDefinition(session)
        self.hub_and_spoke = ConfigurationPolicyHubAndSpokeDefinition(session)
        self.mesh = ConfigurationPolicyMeshDefinition(session)
        self.acl = ConfigurationPolicyAclDefinition(session)
        self.acl_ipv6 = ConfigurationPolicyAclIPv6Definition(session)
        self.device_access = ConfigurationPolicyDeviceAccessDefinition(session)
        self.device_access_ipv6 = ConfigurationPolicyDeviceAccessIPv6Definition(session)


class ConfigurationPolicyContainer:
    def __init__(self, session: ManagerSession):
        self.list = ConfigurationPolicyListContainer(session)
        self.definition = ConfigurationPolicyDefinitionContainer(session)
        self.vsmart_template = ConfigurationVSmartTemplatePolicy(session)
        self.vedge_template = ConfigurationVEdgeTemplatePolicy(session)
        self.security_template = ConfigurationSecurityTemplatePolicy(session)


class ConfigurationSDWANFeatureProfileContainer:
    def __init__(self, session: ManagerSession):
        self.transport = TransportFeatureProfile(client=session)
        self.system = SystemFeatureProfile(client=session)


class ConfigurationFeatureProfileContainer:
    def __init__(self, session: ManagerSession):
        self.sdwan = ConfigurationSDWANFeatureProfileContainer(session=session)


class ConfigurationContainer:
    def __init__(self, session: ManagerSession):
        self.policy = ConfigurationPolicyContainer(session)
        self.feature_profile = ConfigurationFeatureProfileContainer(session=session)


class TroubleshootingToolsContainer:
    def __init__(self, session: ManagerSession):
        self.device_connectivity = TroubleshootingToolsDeviceConnectivity(session)


class RealTimeMonitoringContainer:
    def __init__(self, session: ManagerSession):
        self.reboot_history = RealTimeMonitoringRebootHistory(session)


class APIEndpointContainter:
    def __init__(self, session: ManagerSession):
        self.administration_user_and_group = AdministrationUserAndGroup(session)
        self.certificate_management_vmanage = CertificateManagementVManage(session)
        self.client = Client(session)
        self.cluster_management = ClusterManagement(session)
        self.configuration = ConfigurationContainer(session)
        self.configuration_dashboard_status = ConfigurationDashboardStatus(session)
        self.configuration_device_actions = ConfigurationDeviceActions(session)
        self.configuration_device_software_update = ConfigurationDeviceSoftwareUpdate(session)
        self.configuration_device_template = ConfigurationDeviceTemplate(session)
        self.configuration_settings = ConfigurationSettings(session)
        self.configuration_software_actions = ConfigurationSoftwareActions(session)
        self.configuration_disaster_recovery = ConfigurationDisasterRecovery(session)
        self.monitoring_device_details = MonitoringDeviceDetails(session)
        self.monitoring_security_policy = MonitoringSecurityPolicy(session)
        self.monitoring_server_info = ServerInfo(session)
        self.monitoring_status = MonitoringStatus(session)
        self.sdavc_cloud_connector = SDAVCCloudConnector(session)
        self.tenant_backup_restore = TenantBackupRestore(session)
        self.tenant_management = TenantManagement(session)
        self.tenant_migration = TenantMigration(session)
        self.configuration_feature_profile = ConfigurationFeatureProfile(session)
        self.configuration_group = ConfigurationGroup(session)
        self.sd_routing_configuration_feature_profile = SDRoutingConfigurationFeatureProfile(session)
        self.configuration_device_inventory = ConfigurationDeviceInventory(session)
        self.troubleshooting_tools = TroubleshootingToolsContainer(session)
        self.misc = MiscellaneousEndpoints(session)
        self.real_time_monitoring = RealTimeMonitoringContainer(session)
        self.certificate_management_device = CertificateManagementDevice(session)
        self.url_monitoring = UrlMonitoring(session)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field, field_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class SecurityPortListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    port: Global[str] = Field(description="Ex: 1 or 1-10 by range. Range 0 to 65530")

    @field_validator("port")
    @classmethod
    def check_port(cls, port: Global[str]):
        value = port.value
        if value.count("-") == 0:
            assert 0 <= int(value) <= 65530
            return port

        start_port, end_port = value.split("-")
        start = int(start_port)
        end = int(end_port)
        assert 0 <= start <= 65530
        assert 0 <= end <= 65530
        assert start < end
        return port


class SecurityPortParcel(_ParcelBase):
    entries: List[SecurityPortListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_port(self, port: str):
        self.entries.append(SecurityPortListEntry(port=as_global(port)))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

from catalystwan.api.task_status_api import Task
from catalystwan.endpoints.configuration_device_inventory import (
    ConfigType,
    ConfigurationDeviceInventory,
    DeviceUnlockPayload,
    GenerateBoostrapConfigurationQueryParams,
    UnlockDeviceDetail,
)
from catalystwan.models.device_inventory import BoostrapConfigurationDetails


class ConfigurationDeviceInventoryAPI:
    def __init__(self, session: ManagerSession):
        self.session = session
        self.endpoint = ConfigurationDeviceInventory(session)

    def unlock(self, device_uuid: str, device_type: str, device_details: list) -> Task:
        """
        Unlocks device from config-group
        """
        devices = []
        for device_detail in device_details:
            unlock_device_detail = UnlockDeviceDetail(
                device_id=device_detail["deviceId"], device_ip=device_detail["deviceIP"]
            )
            devices.append(unlock_device_detail)

        payload = DeviceUnlockPayload(device_type=device_type, devices=devices)

        task_id = self.endpoint.unlock(device_uuid=device_uuid, payload=payload).parentTaskId
        return Task(self.session, task_id=task_id)

    def generate_bootstrap_cfg(
        self,
        device_uuid: UUID,
        configtype: ConfigType = "cloudinit",
        incl_def_root_cert: bool = False,
        version: str = "v1",
    ) -> BoostrapConfigurationDetails:
        """
        Returns handy model of generated bootstrap config
        """
        params = GenerateBoostrapConfigurationQueryParams(
            configtype=configtype, incl_def_root_cert=incl_def_root_cert, version=version
        )
        reponse = self.endpoint.generate_bootstrap_configuration(uuid=device_uuid, params=params)

        return BoostrapConfigurationDetails(bootstrap_config=reponse.bootstrap_config)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import Dict

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, post, view
from catalystwan.utils.session_type import ProviderView


class FeatureToCLIPayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    device_specific_variables: Dict[str, str] = Field(alias="device")
    is_edited: bool = Field(alias="isEdited")
    is_master_edited: bool = Field(alias="isMasterEdited")
    is_RFS_required: bool = Field(alias="isRFSRequired")
    template_id: str = Field(alias="templateId")


class ConfigurationDeviceTemplate(APIEndpoints):
    @view({ProviderView})
    @post("/template/device/config/config/")
    def get_device_configuration_preview(self, payload: FeatureToCLIPayload) -> str:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import Any, Dict, List, Optional

from pydantic import BaseModel


def merge(a, b, path=None):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise Exception(f"Conflict at {'.'.join(path + [str(key)])}")
        else:
            a[key] = b[key]
    return a


class FlattenedDictValue(BaseModel):
    value: Any
    data_path: List[str]


def flatten_dict(original_dict: Dict[str, Any]) -> Dict[str, List[FlattenedDictValue]]:
    """
    Flattens a dictionary.
    Each key corresponds to a list of FlattenedDictValue, allowing us to handle repeated keys in nesting.
    """

    def get_flattened_dict(
        original_dict: Dict[str, Any],
        flattened_dict: Optional[Dict[str, List[FlattenedDictValue]]] = None,
        path: Optional[List[str]] = None,
    ):
        if flattened_dict is None:
            flattened_dict = {}
        if path is None:
            path = []
        for key, value in original_dict.items():
            if isinstance(value, dict):
                get_flattened_dict(value, flattened_dict, path=path + [key])
            else:
                if key not in flattened_dict:
                    flattened_dict[key] = []
                if isinstance(value, list) and all([isinstance(v, dict) for v in value]):
                    flattened_value = FlattenedDictValue(
                        value=[get_flattened_dict(v, {}) for v in value], data_path=path
                    )
                    flattened_dict[key].append(flattened_value)
                else:
                    flattened_dict[key].append(FlattenedDictValue(value=value, data_path=path))
        return flattened_dict

    flattened_dict: Dict[str, List[FlattenedDictValue]] = {}
    get_flattened_dict(original_dict, flattened_dict)
    return flattened_dict



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from catalystwan.api.templates.models.cisco_bfd_model import CiscoBFDModel, Color, ColorType

bfd_model = CiscoBFDModel(  # type: ignore
    template_name="cisco_bfd",
    template_description="na",
    multiplier=100,
    poll_interval=50,
    default_dscp=50,
    color=[
        Color(  # type: ignore
            color=ColorType.BIZ_INTERNET, hello_interval=50, multipler=100, pmtu_discovery=False, dscp=100
        ),
        Color(color=ColorType.SILVER, hello_interval=150, multipler=10, dscp=20),  # type: ignore
    ],
)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable

NetworkType = Literal[
    "broadcast",
    "point-to-point",
    "non-broadcast",
    "point-to-multipoint",
]

AuthenticationType = Literal["message-digest"]

AreaType = Literal[
    "stub",
    "nssa",
]

AdvertiseType = Literal[
    "administrative",
    "on-startup",
]

RedistributeProtocol = Literal[
    "static",
    "connected",
    "bgp",
    "omp",
    "nat",
    "eigrp",
]

MetricType = Literal["type1", "type2"]


class SummaryPrefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ip_address: Optional[Union[Global[str], Variable]] = None
    subnet_mask: Optional[Union[Global[str], Variable]] = None


class SummaryRoute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Optional[SummaryPrefix] = None
    cost: Optional[Union[Global[int], Variable, Default[None]]] = None
    no_advertise: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="noAdvertise", validation_alias="noAdvertise", default=None
    )


class OspfInterfaceParametres(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: Optional[Union[Global[str], Variable]]
    hello_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="helloInterval", validation_alias="helloInterval", default=None
    )
    dead_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="deadInterval", validation_alias="deadInterval", default=None
    )
    retransmit_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="retransmitInterval", validation_alias="retransmitInterval", default=None
    )
    cost: Optional[Union[Global[int], Variable, Default[None]]] = None
    priority: Optional[Union[Global[int], Variable, Default[int]]] = None
    network: Optional[Union[Global[NetworkType], Variable, Default[NetworkType]]] = Default[NetworkType](
        value="broadcast"
    )
    passive_interface: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="passiveInterface", validation_alias="passiveInterface", default=None
    )
    authentication_type: Optional[Union[Global[AuthenticationType], Variable, Default[None]]] = Field(
        serialization_alias="type", validation_alias="type", default=None
    )
    message_digest_key: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="messageDigestKey", validation_alias="messageDigestKey", default=None
    )
    md5: Optional[Union[Global[str], Variable, Default[None]]] = None


class OspfArea(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    area_number: Union[Global[int], Variable] = Field(serialization_alias="aNum", validation_alias="aNum")
    area_type: Optional[Union[Global[AreaType], Default[None]]] = Field(
        serialization_alias="aType", validation_alias="aType", default=None
    )
    no_summary: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="noSummary", validation_alias="noSummary", default=None
    )
    interface: Optional[List[OspfInterfaceParametres]] = None
    range: Optional[List[SummaryRoute]]


class RouterLsa(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ad_type: Global[AdvertiseType] = Field(serialization_alias="adType", validation_alias="adType")
    time: Optional[Union[Global[int], Variable]] = None


class RedistributedRoute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Union[Global[RedistributeProtocol], Variable]
    dia: Optional[Union[Global[bool], Variable, Default[bool]]] = None
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )


class OspfData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    router_id: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="routerId", validation_alias="routerId", default=None
    )
    reference_bandwidth: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="referenceBandwidth", validation_alias="referenceBandwidth", default=None
    )
    rfc1583: Optional[Union[Global[bool], Variable, Default[bool]]] = None
    originate: Optional[Union[Global[bool], Default[bool]]] = None
    always: Optional[Union[Global[bool], Variable, Default[bool]]] = None
    metric: Optional[Union[Global[int], Variable, Default[None]]] = None
    metric_type: Optional[Union[Global[MetricType], Variable, Default[None]]] = Field(
        serialization_alias="metricType", validation_alias="metricType", default=None
    )
    external: Optional[Union[Global[int], Variable, Default[int]]] = None
    inter_area: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="interArea", validation_alias="interArea", default=None
    )
    intra_area: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="intraArea", validation_alias="intraArea", default=None
    )
    delay: Optional[Union[Global[int], Variable, Default[int]]] = None
    initial_hold: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="initialHold", validation_alias="initialHold", default=None
    )
    max_hold: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="maxHold", validation_alias="maxHold", default=None
    )
    redistribute: Optional[List[RedistributedRoute]] = None
    router_lsa: Optional[List[RouterLsa]] = Field(
        serialization_alias="routerLsa", validation_alias="routerLsa", default=None
    )
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )
    area: Optional[List[OspfArea]] = None


class OspfCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: OspfData
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional, cast

from catalystwan.api.task_status_api import Task
from catalystwan.api.versions_utils import DeviceVersions, RepositoryAPI
from catalystwan.endpoints.configuration_device_actions import (
    LxcImageActivatePayload,
    LxcImageDeletePayload,
    LxcImageUpgradePayload,
    LxcInstallInput,
)
from catalystwan.endpoints.configuration_device_inventory import DeviceDetailsResponse
from catalystwan.exceptions import EmptyVersionPayloadError, ImageNotInRepositoryError  # type: ignore
from catalystwan.typed_list import DataSequence
from catalystwan.utils.upgrades_helper import get_install_specification, validate_personality_homogeneity

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class LxcActionAPI:
    """
    API methods for virtual software actions. All methods
    are executable on all device categories.

    Usage example:
    # Create session
    session = create_manager_session(...)

    # Prepare specific device for Lxc image actions
    all_dev = session.api.devices.get()
    dev = all_dev.filter(hostname="pm1044")
    ver = "1.0.0_SV3.1.67.0_XE17.14"

    # Upgrade
    del_id = session.api.lxcsoftware.lxcdelete(dev,ver)
    up_id = session.api.lxcsoftware.lxcupgrade(dev,ver)
    act_id = session.api.lxcsoftware.lxcactivate(dev,ver)

    # Check upgrade status
    TaskAPI(session, software_action_id).wait_for_completed()
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session
        self.repository = RepositoryAPI(self.session)
        self.device_versions = DeviceVersions(self.session)

    def lxcactivate(
        self,
        devices: DataSequence[DeviceDetailsResponse],
        version_to_activate: str,
    ) -> Task:
        """
        Requires that selected devices have already version_to_activate installed(upgraded)

        Args:
            devices (List[DeviceDetailsResponse]): For those devices software will be activated
            version_to_activate (Optional[str]): version to be set as current version

            Notice: Have to pass one of those arguments (version_to_activate, image)

        Raises:
            EmptyVersionPayloadError: If selected version_to_activate or image not detected in available files

        Returns:
            str: Activate software action id
        """
        validate_personality_homogeneity(devices)
        version = cast(str, version_to_activate)

        if not version:
            raise ImageNotInRepositoryError(
                "Based on provided arguments, virtual software version to activate on device(s) cannot be detected."
            )

        payload_devices = self.device_versions.get_lxcactivate_device_list(version, devices)
        for device in payload_devices:
            if not device.install_images[0].version_name:
                raise EmptyVersionPayloadError("lxc activate payload contains entry with empty version field.")

        device_type = get_install_specification(devices.first()).device_type.value
        install_input = LxcInstallInput(v_edge_vpn="0", version_type="vmanage")
        partition_payload = LxcImageActivatePayload(
            action="lxc_activate",
            devices=[dev for dev in payload_devices],
            device_type=device_type,
            input=install_input,
        )

        partition_action = self.session.endpoints.configuration_device_actions.process_lxc_activate(
            payload=partition_payload
        )

        return Task(self.session, partition_action.id)

    def lxcupgrade(
        self,
        devices: DataSequence[DeviceDetailsResponse],
        version_to_upgrade: Optional[str] = "",
        image: Optional[str] = "",
    ) -> Task:
        """
        Requires that selected devices have already version_to_upgrade

        Args:
            devices (List[DeviceDetailsResponse]): For those devices software will be activated
            version_to_upgrade (Optional[str]): version to be set as current version
            image (Optional[str]): software image name in available files

            Notice: Have to pass one of those arguments (version_to_activate, image)

        Raises:
            EmptyVersionPayloadError: If selected version_to_activate or image not detected in available files

        Returns:
            str: Activate software action id
        """
        validate_personality_homogeneity(devices)
        if image and not version_to_upgrade:
            version = cast(str, self.repository.get_image_version(image))
        elif version_to_upgrade and not image:
            version = cast(str, version_to_upgrade)
        else:
            raise ValueError("You can not provide software_image and image version at the same time!")

        if not version:
            raise ImageNotInRepositoryError(
                "Based on provided arguments, virtual software version to activate on device(s) cannot be detected."
            )

        payload_devices = self.device_versions.get_lxcupgrade_device_list(version, devices)
        for device in payload_devices:
            if not device.install_images[0].version_name:
                raise EmptyVersionPayloadError("lxc upgrade payload contains entry with empty version field.")

        device_type = get_install_specification(devices.first()).device_type.value
        install_input = LxcInstallInput(v_edge_vpn="0", version_type="vmanage")
        partition_payload = LxcImageUpgradePayload(
            action="lxc_upgrade", devices=[dev for dev in payload_devices], device_type=device_type, input=install_input
        )

        partition_action = self.session.endpoints.configuration_device_actions.process_lxc_upgrade(
            payload=partition_payload
        )

        return Task(self.session, partition_action.id)

    def lxcdelete(
        self,
        devices: DataSequence[DeviceDetailsResponse],
        version_to_delete: str,
    ) -> Task:
        """
        Requires that selected devices have already version_to_delete

        Args:
            devices (List[DeviceDetailsResponse]): For those devices lxc software will be deleted
            version_to_delete (Optional[str]): version to delete

            Notice: Have to pass one of those arguments (version_to_delete)

        Raises:
            EmptyVersionPayloadError: If selected version_to_activate or image not detected in available files

        Returns:
            str: Delete software action id
        """
        validate_personality_homogeneity(devices)
        version = cast(str, version_to_delete)

        if not version:
            raise ImageNotInRepositoryError(
                "Based on provided arguments, virtual software version to delete on device(s) cannot be detected."
            )

        payload_devices = self.device_versions.get_lxcupgrade_device_list(version, devices)
        for device in payload_devices:
            if not device.install_images[0].version_name:
                raise EmptyVersionPayloadError("lxc device payload contains entry with empty version field.")

        device_type = get_install_specification(devices.first()).device_type.value
        partition_payload = LxcImageDeletePayload(
            action="lxc_delete",
            devices=[dev for dev in payload_devices],
            device_type=device_type,
        )

        partition_action = self.session.endpoints.configuration_device_actions.process_lxc_delete(
            payload=partition_payload
        )

        return Task(self.session, partition_action.id)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import json
import logging
from difflib import Differ
from typing import TYPE_CHECKING

from attr import define  # type: ignore
from ciscoconfparse import CiscoConfParse  # type: ignore
from requests.exceptions import HTTPError

from catalystwan.dataclasses import Device
from catalystwan.exceptions import TemplateTypeError
from catalystwan.utils.device_model import DeviceModel
from catalystwan.utils.template_type import TemplateType

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


@define
class CLITemplate:
    template_name: str
    template_description: str
    device_model: DeviceModel
    config: CiscoConfParse = CiscoConfParse([], factory=True)

    def load(self, session: ManagerSession, id: str) -> CiscoConfParse:
        """Load CLI config from template.
        Args:
            id (str): The template id from which load config.
        Raises:
            TemplateTypeError: wrong template type - CLI required.
        Returns:
            CiscoConfParse: Loaded template.
        """
        endpoint = f"/dataservice/template/device/object/{id}"
        config = session.get_json(endpoint)
        if TemplateType(config["configType"]) == TemplateType.FEATURE:
            raise TemplateTypeError(config["templateName"])
        self.config = CiscoConfParse(config["templateConfiguration"].splitlines())
        return self.config

    def load_running(self, session: ManagerSession, device: Device) -> CiscoConfParse:
        """Load running config from device.

        Args:
            session: logged in API client session
            device: The device from which load config.

        Returns:
            CiscoConfParse: A working configuration on the machine.
        """
        encoded_uuid = device.uuid.replace("/", "%2F")
        endpoint = f"/dataservice/template/config/running/{encoded_uuid}"
        config = session.get_json(endpoint)
        self.config = CiscoConfParse(config["config"].splitlines())
        logger.debug(f"Template loaded from {device.hostname}.")
        return self.config

    def generate_payload(self) -> dict:
        config_str = "\n".join(self.config.ioscfg)
        payload = {
            "templateName": self.template_name,
            "templateDescription": self.template_description,
            "deviceType": self.device_model.value,
            "templateConfiguration": config_str,
            "factoryDefault": False,
            "configType": "file",
        }
        if self.device_model not in [
            DeviceModel.VEDGE,
            DeviceModel.VSMART,
            DeviceModel.VMANAGE,
            DeviceModel.VBOND,
        ]:
            payload["cliType"] = "device"
            payload["draftMode"] = False
        return payload

    def update(self, session: ManagerSession, id: str, config: CiscoConfParse) -> bool:
        """Update an existing cli template.

        Args:
            session: logged in API client session
            id: Template id to update.
            config: Updated config.

        Returns:
            bool: True if update template is successful, otherwise - False.

        """
        self.config = config
        config_str = "\n".join(self.config.ioscfg)
        payload = {
            "templateId": id,
            "templateName": self.template_name,
            "templateDescription": self.template_description,
            "deviceType": self.device_model.value,
            "templateConfiguration": config_str,
            "factoryDefault": False,
            "configType": "file",
            "draftMode": False,
        }
        endpoint = f"/dataservice/template/device/{id}"
        try:
            session.put(url=endpoint, json=payload)
        except HTTPError as error:
            if error.response:
                response = json.loads(error.response.text)["error"]
                logger.error(f'Response message: {response["message"]}')
                logger.error(f'Response details: {response["details"]}')
            else:
                logger.error("Response is None.")
            return False
        logger.info(f"Template with name: {self.template_name} - updated.")
        return True

    @staticmethod
    def compare_template(
        first: CiscoConfParse,
        second: CiscoConfParse,
        full: bool = False,
        debug: bool = False,
    ) -> str:
        """

        Args:
            first: First template for comparison.
            second: Second template for comparison.
            full: Return a full comparison if True, otherwise only the lines that differ.
            debug: Adding debug to the logger. Defaults to False.

        Returns:
            str: The compared templates.

        Code    Meaning
        '- '    line unique to sequence 1
        '+ '    line unique to sequence 2
        '  '    line common to both sequences
        '? '    line not present in either input sequence

        Example:
        >>> a = "!\n  tacacs\n  server 192.168.1.1\n   vpn 2\n   secret-key a\n   auth-port 151\n exit".splitlines()
        >>> b = "!\n  tacacs\n  server 192.168.1.1\n   vpn 3\n   secret-key a\n   auth-port 151\n exit".splitlines()
        >>> a_conf = CiscoConfParse(a)
        >>> b_conf = CiscoConfParse(b)
        >>> compare = TemplateAPI.compare_template(a_conf, b_conf full=True)
        >>> print(compare)
          !
            tacacs
            server 192.168.1.1
        -    vpn 2
        ?        ^
        +    vpn 3
        ?        ^
            secret-key a
            auth-port 151
        exit
        """
        first_n = list(map(lambda x: x + "\n", first.ioscfg))
        second_n = list(map(lambda x: x + "\n", second.ioscfg))
        compare = list(Differ().compare(first_n, second_n))
        if not full:
            compare = [x for x in compare if x[0] in ["?", "-", "+"]]
        if debug:
            logger.debug("".join(compare))
        return "".join(compare)

    def compare_with_running(
        self,
        session,
        template: CiscoConfParse,
        device: Device,
        debug: bool = False,
    ) -> str:
        """The comparison of the config with the one running on the machine.

        Args:
            template: The template to compare.
            device: The device on which to compare config.
            full: Return a full comparison if True, otherwise only the lines that differ.
            debug: Adding debug to the logger. Defaults to False.

        Returns:
            str: The compared templates.

        Example:
        >>> a = "!\n  tacacs\n  server 192.168.1.1\n   vpn 512\n   secret-key a\n   auth-port 151\n exit".splitlines()
        >>> a_conf = CiscoConfParse(a)
        >>> device = DevicesAPI(API_SESSION).get(DeviceField.HOSTNAME, device_name)
        >>> compare = TemplateAPI.compare_template(a_conf, device, full=True)
        >>> print(compare)
        .
        .
        .
            zbfw-udp-idle-time    30
           !
          !
        + !
        +   tacacs
        +   server 192.168.1.1
        +    vpn vpn 512
        +    secret-key a
        +    auth-port 151
        +  exit
          omp
           no shutdown
           ecmp-limit       6
        .
        .
        .
        """
        running_config = self.load_running(session, device)
        return self.compare_template(running_config, template, debug)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, get
from catalystwan.utils.operation_status import OperationStatus


class SubTaskData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    status: str
    status_id: str = Field(serialization_alias="statusId", validation_alias="statusId")
    action: Optional[str] = None
    activity: List[str] = []
    current_activity: Optional[str] = Field(
        default=None, serialization_alias="currentActivity", validation_alias="currentActivity"
    )
    action_config: Optional[Union[str, Dict]] = Field(
        default=None, serialization_alias="actionConfig", validation_alias="actionConfig"
    )
    order: Optional[int] = None
    uuid: Optional[str] = None
    hostname: Optional[str] = Field(default=None, serialization_alias="host-name", validation_alias="host-name")
    site_id: Optional[str] = Field(default=None, serialization_alias="site-id", validation_alias="site-id")


class TaskResult(BaseModel):
    result: bool
    sub_tasks_data: List[SubTaskData]


class RunningTaskData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    details_url: Optional[str] = Field(default=None, serialization_alias="detailsURL", validation_alias="detailsURL")
    user_session_username: Optional[str] = Field(
        default=None, serialization_alias="userSessionUserName", validation_alias="userSessionUserName"
    )
    rid: Optional[int] = Field(default=None, serialization_alias="@rid", validation_alias="@rid")
    tenant_name: Optional[str] = Field(default=None, serialization_alias="tenantName", validation_alias="tenantName")
    process_id: Optional[str] = Field(default=None, serialization_alias="processId", validation_alias="processId")
    name: Optional[str] = None
    tenant_id: Optional[str] = Field(default=None, serialization_alias="tenantId", validation_alias="tenantId")
    user_session_ip: Optional[str] = Field(
        default=None, serialization_alias="userSessionIP", validation_alias="userSessionIP"
    )
    action: Optional[str] = None
    start_time: Optional[int] = Field(default=None, serialization_alias="startTime", validation_alias="startTime")
    end_time: Optional[int] = Field(default=None, serialization_alias="endTime", validation_alias="endTime")
    status: Optional[str] = None


class Validation(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    status_type: Optional[str] = Field(default=None, serialization_alias="statusType", validation_alias="statusType")
    activity: Optional[List[str]] = Field(default=None, serialization_alias="activity", validation_alias="activity")
    vmanage_ip: Optional[str] = Field(default=None, serialization_alias="vmanageIP", validation_alias="vmanageIP")
    system_ip: Optional[str] = Field(default=None, serialization_alias="system-ip", validation_alias="system-ip")
    device_id: Optional[str] = Field(default=None, serialization_alias="deviceID", validation_alias="deviceID")
    uuid: Optional[str] = Field(default=None, serialization_alias="uuid", validation_alias="uuid")
    rid: Optional[int] = Field(default=None, serialization_alias="@rid", validation_alias="@rid")
    status_id: str = Field(default=None, serialization_alias="statusId", validation_alias="statusId")
    process_id: Optional[str] = Field(default=None, serialization_alias="processId", validation_alias="processId")
    action_config: Optional[Union[str, Dict]] = Field(
        default=None, serialization_alias="actionConfig", validation_alias="actionConfig"
    )
    current_activity: Optional[str] = Field(
        default=None, serialization_alias="currentActivity", validation_alias="currentActivity"
    )
    action: Optional[str] = Field(default=None, serialization_alias="action", validation_alias="action")
    start_time: Optional[int] = Field(default=None, serialization_alias="startTime", validation_alias="startTime")
    request_status: Optional[str] = Field(
        default=None, serialization_alias="requestStatus", validation_alias="requestStatus"
    )
    status: OperationStatus = Field(default=None, serialization_alias="status", validation_alias="status")
    order: Optional[int] = Field(default=None, serialization_alias="order", validation_alias="order")


class Summary(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    action: Optional[str] = Field(default=None, serialization_alias="action", validation_alias="action")
    name: Optional[str] = Field(default=None, serialization_alias="name", validation_alias="name")
    details_url: Optional[str] = Field(default=None, serialization_alias="detailsURL", validation_alias="detailsURL")
    start_time: Optional[str] = Field(default=None, serialization_alias="startTime", validation_alias="startTime")
    end_time: Optional[str] = Field(default=None, serialization_alias="endTime", validation_alias="endTime")
    user_session_user_name: Optional[str] = Field(
        default=None, serialization_alias="userSessionUserName", validation_alias="userSessionUserName"
    )
    user_session_ip: Optional[str] = Field(
        default=None, serialization_alias="userSessionIP", validation_alias="userSessionIP"
    )
    tenant_name: Optional[str] = Field(default=None, serialization_alias="tenantName", validation_alias="tenantName")
    total: Optional[int] = Field(default=None, serialization_alias="total", validation_alias="total")
    status: Optional[str] = Field(default=None, serialization_alias="status", validation_alias="status")
    count: Optional[dict] = Field(default=None, serialization_alias="count", validation_alias="count")


class TaskData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    data: List[SubTaskData] = []
    validation: Optional[Validation] = None
    summary: Optional[Summary] = None
    is_cancel_enabled: Optional[bool] = Field(
        default=None, serialization_alias="isCancelEnabled", validation_alias="isCancelEnabled"
    )
    is_parallel_execution_enabled: Optional[bool] = Field(
        default=None, serialization_alias="isParallelExecutionEnabled", validation_alias="isParallelExecutionEnabled"
    )


class TasksData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    running_tasks: List[RunningTaskData] = Field(
        default=[], serialization_alias="runningTasks", validation_alias="runningTasks"
    )


class ConfigurationDashboardStatus(APIEndpoints):
    @get("/device/action/status/{task_id}")
    def find_status(self, task_id: str) -> TaskData:
        ...

    @get("/device/action/status/tasks")
    def find_running_tasks(self) -> TasksData:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global
from catalystwan.models.common import TLOCColor


class ColorEntry(BaseModel):
    color: Global[TLOCColor]


class ColorParcel(_ParcelBase):
    entries: List[ColorEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_color(self, color: TLOCColor):
        self.entries.append(ColorEntry(color=as_global(color, TLOCColor)))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
import datetime

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, get


class WebServerCertificateInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    org_unit: str
    org: str
    location: str
    state: str
    country: str
    company_name: str
    not_before: datetime.datetime = Field(serialization_alias="notBefore", validation_alias="notBefore")
    not_after: datetime.datetime = Field(serialization_alias="notAfter", validation_alias="notAfter")
    certificate_details: str = Field(serialization_alias="certificateDetails", validation_alias="certificateDetails")
    validity: str


class CertificateManagementVManage(APIEndpoints):
    def dump_certificate(self):
        # GET /setting/configuration/webserver/certificate/certificate
        ...

    def get_certificate(self):
        # GET /setting/configuration/webserver/certificate/getcertificate
        ...

    def get_csr(self):
        # POST /setting/configuration/webserver/certificate
        ...

    def import_certificate(self):
        # PUT /setting/configuration/webserver/certificate
        ...

    def rollback(self):
        # GET /setting/configuration/webserver/certificate/rollback
        ...

    @get("/setting/configuration/webserver/certificate")
    def show_info(self) -> WebServerCertificateInfo:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Address, IPv4Network
from typing import Any, List, Literal, Optional, Set, Tuple, Union, overload
from uuid import UUID

from pydantic import ConfigDict, Field
from typing_extensions import Annotated

from catalystwan.models.common import ServiceChainNumber, TLOCColor
from catalystwan.models.policy.lists_entries import EncapType
from catalystwan.models.policy.policy_definition import (
    AppListEntry,
    CFlowDAction,
    CountAction,
    DefinitionWithSequencesCommonBase,
    DestinationDataIPv6PrefixListEntry,
    DestinationDataPrefixListEntry,
    DestinationIPEntry,
    DestinationPortEntry,
    DestinationRegionEntry,
    DNSAppListEntry,
    DNSEntry,
    DNSTypeEntryType,
    DREOptimizationAction,
    DSCPEntry,
    FallBackToRoutingAction,
    ForwardingClassEntry,
    LocalTLOCListEntry,
    LocalTLOCListEntryValue,
    LogAction,
    LossProtectionAction,
    LossProtectionFECAction,
    LossProtectionPacketDuplicationAction,
    LossProtectionType,
    Match,
    NATAction,
    NextHopEntry,
    NextHopLooseEntry,
    PacketLengthEntry,
    PLPEntry,
    PolicerListEntry,
    PolicyActionType,
    PolicyDefinitionBase,
    PolicyDefinitionSequenceBase,
    PrefferedColorGroupListEntry,
    ProtocolEntry,
    RedirectDNSAction,
    SecureInternetGatewayAction,
    ServiceChainEntry,
    ServiceChainEntryValue,
    ServiceNodeGroupAction,
    SourceDataIPv6PrefixListEntry,
    SourceDataPrefixListEntry,
    SourceIPEntry,
    SourcePortEntry,
    TCPEntry,
    TCPOptimizationAction,
    TLOCEntry,
    TLOCEntryValue,
    TLOCListEntry,
    TrafficToEntry,
    VPNEntry,
    accept_action,
)

TrafficDataPolicySequenceEntry = Annotated[
    Union[
        PacketLengthEntry,
        PLPEntry,
        ProtocolEntry,
        DSCPEntry,
        SourceIPEntry,
        SourcePortEntry,
        DestinationIPEntry,
        DestinationPortEntry,
        TCPEntry,
        DNSEntry,
        TrafficToEntry,
        SourceDataPrefixListEntry,
        DestinationDataPrefixListEntry,
        SourceDataIPv6PrefixListEntry,
        DestinationDataIPv6PrefixListEntry,
        DestinationRegionEntry,
        DNSAppListEntry,
        AppListEntry,
    ],
    Field(discriminator="field"),
]

TrafficDataPolicySequenceActions = Any  # TODO


class TrafficDataPolicyHeader(PolicyDefinitionBase):
    type: Literal["data"] = "data"


class TrafficDataPolicySequenceMatch(Match):
    entries: List[TrafficDataPolicySequenceEntry] = []


class TrafficDataPolicySequence(PolicyDefinitionSequenceBase):
    sequence_type: Literal["data"] = Field(
        default="data", serialization_alias="sequenceType", validation_alias="sequenceType"
    )
    match: TrafficDataPolicySequenceMatch = TrafficDataPolicySequenceMatch()
    actions: List[TrafficDataPolicySequenceActions] = []
    model_config = ConfigDict(populate_by_name=True)

    def match_app_list(self, app_list_id: UUID) -> None:
        self._insert_match(AppListEntry(ref=app_list_id))

    def match_dns_app_list(self, dns_app_list_id: UUID) -> None:
        self._insert_match(DNSAppListEntry(ref=dns_app_list_id))

    def match_dns_request(self) -> None:
        self._insert_match(DNSEntry(value="request"))

    def match_dns_response(self) -> None:
        self._insert_match(DNSEntry(value="response"))

    def match_dscp(self, dscp: int) -> None:
        self._insert_match(DSCPEntry(value=str(dscp)))

    def match_packet_length(self, packet_lengths: Tuple[int, int]) -> None:
        self._insert_match(PacketLengthEntry.from_range(packet_lengths))

    def match_low_plp(self) -> None:
        self._insert_match(PLPEntry(value="low"))

    def match_high_plp(self) -> None:
        self._insert_match(PLPEntry(value="high"))

    def match_protocols(self, protocols: Set[int]) -> None:
        self._insert_match(ProtocolEntry.from_protocol_set(protocols))

    def match_source_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(SourceDataPrefixListEntry(ref=data_prefix_list_id))

    def match_source_ip(self, networks: List[IPv4Network]) -> None:
        self._insert_match(SourceIPEntry.from_ipv4_networks(networks))

    def match_source_port(self, ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> None:
        self._insert_match(SourcePortEntry.from_port_set_and_ranges(ports, port_ranges))

    def match_destination_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(DestinationDataPrefixListEntry(ref=data_prefix_list_id))

    def match_destination_ip(self, networks: List[IPv4Network]) -> None:
        self._insert_match(DestinationIPEntry.from_ipv4_networks(networks))

    def match_primary_destination_region(self) -> None:
        self._insert_match(DestinationRegionEntry(value="primary-region"))

    def match_secondary_destination_region(self) -> None:
        self._insert_match(DestinationRegionEntry(value="secondary-region"))

    def match_other_destination_region(self) -> None:
        self._insert_match(DestinationRegionEntry(value="other-region"))

    def match_destination_port(self, ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> None:
        self._insert_match(DestinationPortEntry.from_port_set_and_ranges(ports, port_ranges))

    def match_tcp(self) -> None:
        self._insert_match(TCPEntry())

    def match_traffic_to_access(self) -> None:
        self._insert_match(TrafficToEntry(value="access"))

    def match_traffic_to_core(self) -> None:
        self._insert_match(TrafficToEntry(value="core"))

    def match_traffic_to_service(self) -> None:
        self._insert_match(TrafficToEntry(value="service"))

    def associate_count_action(self, counter_name: str) -> None:
        self._insert_action(CountAction(parameter=counter_name))

    def associate_log_action(self) -> None:
        self._insert_action(LogAction())

    @accept_action
    def associate_dscp_action(self, dscp: int) -> None:
        self._insert_action_in_set(DSCPEntry(value=str(dscp)))

    @accept_action
    def associate_forwarding_class_action(self, fwclass: str) -> None:
        self._insert_action_in_set(ForwardingClassEntry(value=fwclass))

    @accept_action
    def associate_local_tloc_action(self, color: TLOCColor, encap: EncapType, restrict: bool = False) -> None:
        tloc_entry = LocalTLOCListEntry(
            value=LocalTLOCListEntryValue(
                color=color,
                encap=encap,
                restrict="" if restrict else None,
            )
        )
        self._insert_action_in_set(tloc_entry)

    @accept_action
    def associate_preffered_color_group(self, color_group_list_id: UUID, restrict: bool = False) -> None:
        self._insert_action_in_set(PrefferedColorGroupListEntry(ref=color_group_list_id, color_restrict=restrict))

    @accept_action
    def associate_cflowd_action(self) -> None:
        self._insert_action(CFlowDAction())

    @overload
    def associate_nat_action(self, *, nat_pool: int) -> None:
        ...

    @overload
    def associate_nat_action(self, *, vpn_fallback: bool = False, vpn: int = 0) -> None:
        ...

    @accept_action
    def associate_nat_action(self, *, nat_pool: Optional[int] = None, vpn_fallback: bool = False, vpn: int = 0) -> None:
        if nat_pool:
            nat_action = NATAction.from_nat_pool(nat_pool=nat_pool)
        else:
            nat_action = NATAction.from_nat_vpn(fallback=vpn_fallback, vpn=vpn)
        self._insert_action(nat_action)

    @accept_action
    def associate_next_hop_action(self, next_hop: IPv4Address, loose: bool = False) -> None:
        self._insert_action_in_set(NextHopEntry(value=next_hop))
        self._insert_action_in_set(NextHopLooseEntry(value="true" if loose else "false"))

    @accept_action
    def associate_policer_list_action(self, policer_list_id: UUID) -> None:
        self._insert_action_in_set(PolicerListEntry(ref=policer_list_id))

    @overload
    def associate_redirect_dns_action(self, *, ip: IPv4Address) -> None:
        ...

    @overload
    def associate_redirect_dns_action(self, *, dns_type: DNSTypeEntryType = "host") -> None:
        ...

    @accept_action
    def associate_redirect_dns_action(self, *, ip=None, dns_type=None) -> None:
        if ip:
            redirect_dns_action = RedirectDNSAction.from_ip_address(ip)
        else:
            redirect_dns_action = RedirectDNSAction.from_dns_type(dns_type)
        self._insert_action(redirect_dns_action)

    @accept_action
    def associate_local_service_chain_action(
        self, sc_type: ServiceChainNumber, vpn: int, restrict: bool = False
    ) -> None:
        self._insert_action_in_set(
            ServiceChainEntry(
                value=ServiceChainEntryValue(
                    type=sc_type,
                    vpn=str(vpn),
                    restrict="" if restrict else None,
                    local="",
                )
            )
        )

    @accept_action
    def associate_remote_service_chain_action(
        self,
        sc_type: ServiceChainNumber,
        vpn: int,
        ip: IPv4Address,
        color: TLOCColor,
        encap: EncapType,
        restrict: bool = False,
    ) -> None:
        self._insert_action_in_set(
            ServiceChainEntry(
                value=ServiceChainEntryValue(
                    type=sc_type,
                    vpn=str(vpn),
                    restrict="" if restrict else None,
                    tloc=TLOCEntryValue(
                        ip=ip,
                        color=color,
                        encap=encap,
                    ),
                )
            )
        )

    @accept_action
    def associate_app_qoe_optimization_action(
        self, tcp: bool = False, dre: bool = False, service_node_group: Optional[str] = None
    ) -> None:
        if tcp:
            self._insert_action(TCPOptimizationAction())
        else:
            self._remove_action(TCPOptimizationAction().type)
        if dre:
            self._insert_action(DREOptimizationAction())
        else:
            self._remove_action(DREOptimizationAction().type)
        if service_node_group is not None:
            self._insert_action(ServiceNodeGroupAction(parameter=service_node_group))
        else:
            self._remove_action(ServiceNodeGroupAction().type)

    @accept_action
    def associate_loss_correction_fec_action(self, adaptive: bool = False, threshold: Optional[int] = None) -> None:
        self._remove_action(LossProtectionPacketDuplicationAction().type)
        fec_type: LossProtectionType = "fecAdaptive" if adaptive else "fecAlways"
        fec_value = str(threshold) if adaptive and threshold is not None else None
        self._insert_action(LossProtectionAction(parameter=fec_type))
        self._insert_action(LossProtectionFECAction(parameter=fec_type, value=fec_value))

    @accept_action
    def associate_loss_correction_packet_duplication_action(self) -> None:
        self._remove_action(LossProtectionFECAction().type)
        self._insert_action(LossProtectionAction(parameter="packetDuplication"))
        self._insert_action(LossProtectionPacketDuplicationAction())

    @accept_action
    def associate_vpn_action(self, vpn: int) -> None:
        # TLOC or Next Hop is mandatory when configuring VPN. Please populate Action > TLOC or Action > Next Hop
        self._insert_action_in_set(VPNEntry(value=str(vpn)))

    @overload
    def associate_tloc_action(self, *, tloc_list_id: UUID) -> None:
        ...

    @overload
    def associate_tloc_action(self, *, ip: IPv4Address, color: TLOCColor, encap: EncapType) -> None:
        ...

    @accept_action
    def associate_tloc_action(self, *, tloc_list_id=None, ip=None, color=None, encap=None) -> None:
        # VPN is mandatory when configuring TLOC. Please populate Action > VPN.
        if tloc_list_id is not None:
            self._insert_action_in_set(TLOCListEntry(ref=tloc_list_id))
        else:
            self._insert_action_in_set(
                TLOCEntry(
                    value=TLOCEntryValue(
                        ip=ip,
                        color=color,
                        encap=encap,
                    )
                )
            )

    @accept_action
    def associate_secure_internet_gateway_action(self, fallback_to_routing: bool = False) -> None:
        # Secure Internet Gateway cannot be enabled with NAT Pool or NAT VPN or Next Hop.
        self._insert_action(SecureInternetGatewayAction())
        if fallback_to_routing:
            self._insert_action(FallBackToRoutingAction())
        else:
            self._remove_action(FallBackToRoutingAction().type)


class TrafficDataPolicy(TrafficDataPolicyHeader, DefinitionWithSequencesCommonBase):
    sequences: List[TrafficDataPolicySequence] = []
    model_config = ConfigDict(populate_by_name=True)

    def add_ipv4_sequence(
        self, name: str = "Custom", base_action: PolicyActionType = "drop", log: bool = False
    ) -> TrafficDataPolicySequence:
        seq = TrafficDataPolicySequence(
            sequence_name=name,
            base_action=base_action,
            sequence_ip_type="ipv4",
        )
        self.add(seq)
        return seq



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import Union

from pydantic import BaseModel
from requests import HTTPError, RequestException


class ManagerErrorInfo(BaseModel):
    message: Union[str, None] = None
    details: Union[str, None] = None
    code: Union[str, None] = None


class CatalystwanException(Exception):
    """Superclass of all catalystwan SDK exception types."""


class ManagerRequestException(RequestException, CatalystwanException):
    """Exception raised when there is ambigous problem during sending of request to Manager"""


class ManagerHTTPError(HTTPError, ManagerRequestException):
    def __init__(self, *args, error_info: ManagerErrorInfo, **kwargs):
        """Initialize RequestException with `error_info`, `request` and `response` objects."""
        self.info = error_info
        info_str = str(self.info)
        _args = args
        if not _args:
            _args = (info_str,)
        else:
            _args = (str(_args[0]) + "\n" + info_str,) + _args[1:]
        super().__init__(*_args, **kwargs)


class DefaultPasswordError(CatalystwanException):
    """Default password for SDWAN Manager user was detected and needs to be changed."""

    pass


class InvalidOperationError(CatalystwanException):
    """The exception that is thrown when a method call is invalid for the object's current state."""

    pass


class RetrieveIntervalOutOfRange(CatalystwanException):
    pass


class VersionDeclarationError(CatalystwanException):
    """

    The exception that is thrown in one of two below cases.
    1. User passes software image and version, at the same time.
    2. User doesn't passes any of them.
    """

    pass


class ImageNotInRepositoryError(CatalystwanException):
    """The exception that is thrown, if image is not in vManage images Repository"""

    pass


class EmptyVersionPayloadError(CatalystwanException):
    """Used when a version is not found in device available or current versions."""

    pass


class TemplateNotFoundError(CatalystwanException):
    """Used when a template item is not found."""

    def __init__(self, template):
        self.message = f"No such template: '{template}'"


class AttachedError(CatalystwanException):
    """Used when delete attached template."""

    def __init__(self, template):
        self.message = f"Template: {template} is attached to device."


class TemplateTypeError(CatalystwanException):
    """Used when wrong type template."""

    def __init__(self, name):
        self.message = f"Template: {name} - wrong template type."


class AlreadyExistsError(CatalystwanException):
    """Raised when an entity that we attempted to create already exists."""

    pass


class APIVersionError(CatalystwanException):
    """Raised when API is unsupported in running vManage version."""

    def __init__(self, item, supported, current):
        self.message = f"vManage is running: {current} but {item} only supported in API version: {supported}"


class APIViewError(CatalystwanException):
    """Raised when API is not allowed for given session type / view."""

    def __init__(self, item, allowed, current):
        self.message = f"Current view is: {current} but {item} only allowed for views: {allowed}"


class APIRequestPayloadTypeError(CatalystwanException):
    """Raised when unsupported payload type is passed to vManage request."""

    def __init__(self, item):
        self.message = f"Unsupported payload type: {type(item)} for vManage request"


class EmptyTaskResponseError(CatalystwanException):
    """Raised if task is registred by vManage, but reponse content is empty"""

    pass


class TaskNotRegisteredError(CatalystwanException):
    """Raised if task_id is generated, but it's not registere in vManage"""

    pass


class TaskValidationError(CatalystwanException):
    """Raised if task has not been validated"""

    pass


class MultiplePersonalityError(CatalystwanException):
    """Raised if Device DataSequnce contains devices with multiples personalities"""


class SessionNotCreatedError(CatalystwanException):
    """Raised when vManage session cannot be created"""

    pass


class TenantSubdomainNotFound(CatalystwanException):
    """Raised when given subdomain does not exist"""

    pass


class APIEndpointError(CatalystwanException):
    """Raised when there is a problem with endpoint definition"""

    pass


class TenantMigrationExportFileNotFound(CatalystwanException):
    """Raised when client cannot find exported file name in export task result"""

    pass


class TenantMigrationPreconditionsError(CatalystwanException):
    """Raised when preconditions for tenant migration fail"""

    pass


class ManagerReadyTimeout(CatalystwanException):
    """Raised when waiting for server ready flag took longer than expected"""

    pass


class CatalystwanDeprecationWarning(DeprecationWarning):
    """Warning issued when using deprecated features or functionality in the Catalystwan SDK.

    This warning indicates that the current usage of certain features or functionality within the Catalystwan SDK
    is deprecated and may be removed in future versions. It serves as a notice to developers to update their code
    to use the recommended alternatives.
    """

    message: str

    def __init__(self, message) -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        message = (
            f"{self.message}. Deprecated in catalystwan (0.y.z), will be removed in (1.y.z)."
            f" Major version zero (0.y.z) is for initial development. Anything MAY change at any time."
            f" The public API SHOULD NOT be considered stable."
        )
        return message



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from pathlib import Path
from unittest.mock import patch

from parameterized import parameterized  # type: ignore

from catalystwan.api.logs_api import LogsAPI


class TestLogsAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.logs = [
            {"logid": "12345", "logmessage": "message from log 1", "tenant": "tenant234", "entry_time": 1671028426000},
            {"logid": "67890", "logmessage": "message from log 2", "tenant": "tenant987", "entry_time": 1671028426000},
        ]

    @patch("catalystwan.session.ManagerSession")
    def test_get_auditlogs_file_path_not_provided(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.logs
        default_file_path = Path(__file__).parents[1].joinpath("api").joinpath("audit.log")
        # Act
        LogsAPI(mock_session).get_auditlogs()
        does_file_exist = default_file_path.is_file()
        # Assert
        self.assertTrue(does_file_exist)

    @parameterized.expand([["test_file.log"], ["other_test_file.log"]])
    @patch("catalystwan.session.ManagerSession")
    def test_get_auditlogs_file_path(self, file_name, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.logs
        file_path = Path(__file__).parents[2].joinpath(file_name)
        # Act
        LogsAPI(mock_session).get_auditlogs(file_path=file_name)
        does_file_exist = file_path.is_file()
        # Assert
        self.assertTrue(does_file_exist)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from catalystwan.dataclasses import TenantAAA, TenantRadiusServer, TenantTacacsServer
from catalystwan.exceptions import CatalystwanException
from catalystwan.utils.creation_tools import asdict, create_dataclass

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

logger = logging.getLogger(__name__)


class AAAConfigNotPresent(CatalystwanException):
    pass


def status_ok(func):
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        return True if response.status_code in [200, 204] else False

    return wrapper


class TenantAaaAPI:
    """
    Used to configure  mtt tenant management users remote servers
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session
        self.url_path = "/dataservice/admin/aaa"

    def __str__(self) -> str:
        return str(self.session)

    @property
    def tenant_id(self):
        return self.session.get_tenant_id()

    def aaa_exists(self) -> bool:
        return True if self.session.get_data(self.url_path) else False

    @status_ok
    def add_aaa(self, tenant_aaa: TenantAAA):
        """ "
        TenantAAA:
            "authOrder": [ "local", "radius", "tacacs"],
            "authFallback": true,
            "adminAuthOrder": false,
            "auditDisable": false,
            "accounting": false,
            "radiusServers": "server1"
        returns bool depending on the api post call
        """
        data = asdict(tenant_aaa)  # type: ignore
        return self.session.post(url=self.url_path, json=data)

    def get_aaa(self) -> TenantAAA:
        """
        Returns the Tenant AAA
        :param aaa:
        :return:
        """
        logger.debug(f"AAA config {self.tenant_id}.")
        tenant_aaa = self.session.get_data(self.url_path)
        # return tenant_aaa
        return create_dataclass(TenantAAA, tenant_aaa)

    @status_ok
    def del_aaa(self):
        """
        Delete aaa works only for tenants
        :return:
        """
        if not self.aaa_exists():
            raise AAAConfigNotPresent(f"No AAA config present for Tenant id={self.tenant_id}")
        logger.debug(f"Delete AAA config on tenant_id={self.tenant_id}.")
        return self.session.delete(self.url_path)

    @status_ok
    def put_aaa(self, tenant_AAA: TenantAAA):
        """
        Updated the AAA for tenant
        :return:
        """
        data = asdict(tenant_AAA)  # type: ignore
        return self.session.put(url=self.url_path, json=data)


class TenantRadiusAPI:
    """
    Used to configure  mtt tenant remote aaa radius servers
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session
        self.url_path = "/dataservice/admin/radius"
        self.tenant_id = self.session.get_tenant_id()

    def __str__(self) -> str:
        return str(self.session)

    @status_ok
    def add_radius(self, radius_server: TenantRadiusServer):
        """
        Create RADIUS for tenant
        :param radius_server:
        :return:
        """
        logger.debug(f"Add RADIUS config tenant_id={self.tenant_id}.")
        data = asdict(radius_server)  # type: ignore
        return self.session.post(url=self.url_path, json=data)

    @status_ok
    def put_radius(self, radius_server: TenantRadiusServer):
        """
        edit radius server
        :param radius_server:
        :return:
        """
        logger.debug(f"Update RADIUS config tenant_id={self.tenant_id}.")
        data = asdict(radius_server)  # type: ignore
        return self.session.put(url=self.url_path, json=data)

    @status_ok
    def delete_radius(self):
        """
        edit radius server
        :param radius_server:
        :return: True|False
        """
        logger.debug(f"Delete RADIUS config tenant_id={self.tenant_id}.")
        return self.session.delete(self.url_path)

    def get_radius(self) -> TenantRadiusServer:
        """
        Retrieve Radius server
        :return: TenantRadiusServer
        """
        logger.debug(f"RADIUS config tenant_id={self.tenant_id}.")
        data = self.session.get_data(self.url_path)
        return create_dataclass(TenantRadiusServer, data)


class TenantTacacsAPI:
    """
    Used to configure mtt tenant remote aaa TACACS servers
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session
        self.url_path = "/dataservice/admin/tacacs"
        self.tenant_id = self.session.get_tenant_id()

    def __str__(self) -> str:
        return str(self.session)

    @status_ok
    def add_tacacs(self, tacacs_server: TenantTacacsServer):
        """
        Create TACACS for tenant
        :param tacacs_server:
        :return:
        """
        logger.debug(f"TACACS config tenant_id={self.tenant_id}.")
        data = asdict(tacacs_server)  # type: ignore
        return self.session.post(url=self.url_path, json=data)

    @status_ok
    def put_tacacs(self, tacacs_server: TenantTacacsServer):
        """
        Update tacacs server
        :param tacacs_server:
        :return:
        """
        logger.debug(f"Update TACACS config tenant_id={self.tenant_id}.")
        data = asdict(tacacs_server)  # type: ignore
        return self.session.put(url=self.url_path, json=data)

    @status_ok
    def delete_tacacs(self):
        """
        Deletes tacacs server
        :param tacacs_server:
        :return: True|False
        """
        logger.debug(f"Delete TACACS config tenant_id={self.tenant_id}.")
        return self.session.delete(self.url_path)

    def get_tacacs(self) -> TenantTacacsServer:
        """
        Retrieves Tacacs server
        :return: TenantTacacsServer
        """
        logger.debug(f"TACACS config tenant_id={self.tenant_id}.")
        data = self.session.get_data(self.url_path)
        return create_dataclass(TenantTacacsServer, data)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from packaging.version import Version  # type: ignore

from catalystwan.api.task_status_api import Task, TaskResult
from catalystwan.endpoints.tenant_migration import ImportInfo, MigrationFile, MigrationTokenQueryParams
from catalystwan.exceptions import TenantMigrationExportFileNotFound
from catalystwan.models.tenant import TenantExport

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class ExportTask(Task):
    def __init__(self, session: ManagerSession, task_id: str):
        super().__init__(session, task_id)

    @staticmethod
    def get_file_name_from_export_task_result(task_result: TaskResult) -> str:
        filepath = re.search("""file location: (.*)""", task_result.sub_tasks_data[0].activity[-1])
        if not filepath:
            raise TenantMigrationExportFileNotFound("Client cannot find export file name in export task result")
        return Path(filepath.group(1)).name

    def wait_for_file(self) -> str:
        task_result = super().wait_for_completed()
        return self.get_file_name_from_export_task_result(task_result)


class ImportTask(Task):
    def __init__(self, session: ManagerSession, import_info: ImportInfo):
        super().__init__(session, import_info.process_id)
        self.import_info = import_info


class TenantMigrationAPI:
    """Set of methods used in tenant migration"""

    def __init__(self, session: ManagerSession):
        self.session = session

    def export_tenant(self, tenant: TenantExport) -> ExportTask:
        """Exports the deployment and configuration data from a Cisco vManage instance.
        Should be executed on migration origin.

        Args:
            tenant (Tenant): Tenant object containig required fields: desc, name, subdomain, org_name

        Returns:
            Task: object representing initiated export process
        """
        process_id = self.session.endpoints.tenant_migration.export_tenant_data(tenant).process_id
        return ExportTask(self.session, process_id)

    def download(self, download_path: Path, remote_filename: str = "default.tar.gz"):
        """Download exported deployment and configuration data from a Cisco vManage instance
        to a local file system. Should be executed on migration origin.

        Args:
            download_path (Path): full download path containing a filename eg.: Path("/home/user/tenant-export.tar.gz")
            remote_filename (str): path to exported tenant migration file on vManage
        """
        tenant_data = self.session.endpoints.tenant_migration.download_tenant_data(remote_filename)
        with open(download_path, "wb") as file:
            file.write(tenant_data)

    def import_tenant(self, import_file: Path, migration_key: Optional[str] = None) -> ImportTask:
        """Imports the deployment and configuration data into multi-tenant vManage instance.
        Should be executed on migration target.

        Args:
            import_file (Path): full path to previously exported data file
            migration_key (str): migration key (required starting from 20.13)

        Returns:
            ImportTask: object representing initiated import process
        """
        import_file_payload = MigrationFile(import_file)
        if self.session.api_version >= Version("20.13") and migration_key is not None:
            import_info = self.session.endpoints.tenant_migration.import_tenant_data_with_key(
                import_file_payload, migration_key
            )
        else:
            import_info = self.session.endpoints.tenant_migration.import_tenant_data(import_file_payload)
        return ImportTask(self.session, import_info)

    def store_token(self, migration_id: str, download_path: Path):
        """Stores migration token as text file for given migration identifier.
        Should be executed on migration target.

        Args:
            migration_id (str): migration identifier (it is obtained after import tenant task is finished)
            download_path (Path): full download path containing a filename eg.: Path("/home/user/import-token.txt")
        """
        params = MigrationTokenQueryParams(migration_id=migration_id)
        token = self.session.endpoints.tenant_migration.get_migration_token(params)
        with open(download_path, "w") as file:
            file.write(token)

    def migrate_network(self, token_file: Path) -> Task:
        """Starts migration procedure on migration origin.

        Args:
            token_file (Path): full path to previously stored text file with import token.

        Returns:
            Task: object representing initiated migration process
        """
        with open(token_file, "r") as file:
            token = file.read()
        process_id = self.session.endpoints.tenant_migration.migrate_network(token).process_id
        return Task(self.session, process_id)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# type: ignore
from catalystwan.api.templates.models.cisco_aaa_model import (
    CiscoAAAModel,
    DomainStripping,
    RadiusGroup,
    RadiusServer,
    TacacsGroup,
    TacacsServer,
    User,
)
from catalystwan.utils.device_model import DeviceModel

users = [
    User(name="admin", password="str", secret="zyx", privilege="15"),  # pragma: allowlist secret
    User(name="user", password="rnd", secret="dnr", privilege="14"),  # pragma: allowlist secret
]

# CiscoAAAModel(domain-stripping="?")
cisco_aaa = CiscoAAAModel(
    template_name="iuo",
    template_description="zyx",
    device_models=[DeviceModel.VEDGE_C8000V],
    user=users,
    authentication_group=True,
    accounting_group=False,
    radius=[
        RadiusGroup(  # type: ignore
            group_name="xyz", vpn=1, source_interface="GIG11", server=[RadiusServer(address="1.1.1.1", key="21")]
        )
    ],
    domain_stripping=DomainStripping.NO,  # type: ignore
)

cisco_aaa_device_specific = CiscoAAAModel(
    template_name="cisco_aaa_device_specific",
    template_description="caaadp",
    device_models=[DeviceModel.VEDGE_C8000V],
    user=users,
    authentication_group=True,
    accounting_group=False,
    radius=[
        RadiusGroup(  # type: ignore
            group_name="xyz", vpn=1, source_interface="GIG11", server=[RadiusServer(address="1.1.1.1", key="21")]
        )
    ],
    domain_stripping=DomainStripping.NO,
)

complex_aaa_model = CiscoAAAModel(
    template_name="complex_aaa",
    template_description="na",
    user=[
        User(name="test1", password="*****", secret="secret", privilege="1"),  # pragma: allowlist secret
        User(name="test2", password="*****", secret="secret", privilege="15"),  # pragma: allowlist secret
    ],
    authentication_group=True,
    accounting_group=False,
    radius=[
        RadiusGroup(
            group_name="group1",
            vpn=10,
            source_interface="Gig1",
            server=[
                RadiusServer(address="1.1.1.1", key="test_key", secret_key="secret_key")  # pragma: allowlist secret
            ],
        ),
        RadiusGroup(
            group_name="group2",
            vpn=11,
            source_interface="Gig2",
            server=[
                RadiusServer(
                    address="1.1.2.1",
                    key="test_key2",
                    secret_key="secret_key2",  # pragma: allowlist secret
                )
            ],
        ),
    ],
    domain_stripping=DomainStripping.RIGHT_TO_LEFT,
    tacacs=[
        TacacsGroup(
            group_name="group1",
            vpn=0,
            source_interface="Gig0",
            server=[TacacsServer(address="1.1.1.1", key="key", secret_key="secret_key")],  # pragma: allowlist secret
        )
    ],
)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from catalystwan.endpoints import APIEndpointClient, APIEndpoints, get
from catalystwan.models.misc.application_protocols import ApplicationProtocolMap
from catalystwan.typed_list import DataSequence


class MiscellaneousEndpoints(APIEndpoints):
    def __init__(self, client: APIEndpointClient):
        self._client = client
        self._basepath = ""

    @get("/app/json/application_protocol.json", "data")
    def get_application_protocols(self) -> DataSequence[ApplicationProtocolMap]:
        """Not in spec, provides protocol name to protocol/port number mapping

        Returns:
            DataSequence[ApplicationProtocolMap]
        """
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import BaseModel


class SecurityPolicyDeviceList(BaseModel):
    amp_down: List[str]
    amp_up: List[str]
    ips_down: List[str]
    ips_up: List[str]
    urlf_down: List[str]
    urlf_up: List[str]
    zbfw_down: List[str]
    zbfw_up: List[str]



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Network
from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class SecurityDataPrefixEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    ip_prefix: Global[IPv4Network] = Field(serialization_alias="ipPrefix", validation_alias="ipPrefix")


class SecurityDataPrefixParcel(_ParcelBase):
    entries: List[SecurityDataPrefixEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_prefix(self, ip_prefix: IPv4Network):
        self.entries.append(SecurityDataPrefixEntry(ip_prefix=as_global(ip_prefix)))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import List

from pydantic import BaseModel

from catalystwan.endpoints import APIEndpoints, get, view
from catalystwan.utils.session_type import ProviderAsTenantView, TenantView


class BackupFiles(BaseModel):
    backup_files: List[str]


class TenantBackupRestore(APIEndpoints):
    @view({ProviderAsTenantView})
    def delete_tenant_backup(self):
        # DELETE /tenantbackup/delete
        ...

    @view({ProviderAsTenantView, TenantView})
    def download_existing_backup_file(self):
        # GET /tenantbackup/download/{path}
        ...

    @view({ProviderAsTenantView, TenantView})
    def export_tenant_backup(self):
        # GET /tenantbackup/export
        ...

    @view({ProviderAsTenantView})
    def import_tenant_backup(self):
        # POST /tenantbackup/import
        ...

    @view({ProviderAsTenantView, TenantView})
    @get("/tenantbackup/list")
    def list_tenant_backup(self) -> BackupFiles:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Union, cast

from jinja2 import DebugUndefined, Environment, FileSystemLoader, meta  # type: ignore
from pydantic import BaseModel, model_validator

from catalystwan.api.templates.device_variable import DeviceVariable
from catalystwan.utils.device_model import DeviceModel
from catalystwan.utils.dict import FlattenedDictValue, flatten_dict
from catalystwan.utils.feature_template.find_template_values import find_template_values
from catalystwan.utils.pydantic_field import get_extra_field

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class FeatureTemplateValidator(BaseModel, ABC):
    @model_validator(mode="before")
    @classmethod
    def map_fields(cls, values: Union[Any, Dict[str, Union[List[FlattenedDictValue], Any]]]):
        if not isinstance(values, dict):
            return values
        for field_name, field_info in cls.model_fields.items():
            vmanage_key = get_extra_field(field_info, "vmanage_key")
            if vmanage_key in values:
                payload_name = vmanage_key
            elif field_info.alias in values:
                payload_name = field_info.alias
            elif field_name in values:
                payload_name = field_name
            else:
                continue
            data_path = get_extra_field(field_info, "data_path", [])
            value = values.pop(payload_name)
            if value and isinstance(value, list) and all([isinstance(v, FlattenedDictValue) for v in value]):
                for template_value in value:
                    if template_value.data_path == data_path:
                        values[field_name] = template_value.value
                        break
            else:
                values[field_name] = value
        return values


class FeatureTemplate(FeatureTemplateValidator, ABC):
    template_name: str
    template_description: str
    device_models: List[DeviceModel] = []
    device_specific_variables: Dict[str, DeviceVariable] = {}

    def generate_payload(self, session: ManagerSession) -> str:
        env = Environment(
            loader=FileSystemLoader(self.payload_path.parent),
            trim_blocks=True,
            lstrip_blocks=True,
            undefined=DebugUndefined,
        )
        template = env.get_template(self.payload_path.name)
        output = template.render(self.model_dump(mode="json"))

        ast = env.parse(output)
        if meta.find_undeclared_variables(ast):
            print(meta.find_undeclared_variables(ast))
            raise Exception
        return output

    def generate_cli(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def payload_path(self) -> Path:
        raise NotImplementedError()

    @property
    def type(self) -> str:
        raise NotImplementedError()

    @model_validator(mode="before")
    @classmethod
    def remove_device_variables(cls, values):
        if "device_specific_variables" not in values:
            values["device_specific_variables"] = {}
        to_delete = {}

        # TODO: Add support for nested models with DeviceVariable
        for key, value in values.items():
            if isinstance(value, DeviceVariable):
                to_delete[key] = value
                field_key = cls.model_fields[key].json_schema_extra.get("vmanage_key", cls.model_fields[key].alias)
                values["device_specific_variables"][field_key] = DeviceVariable(name=value.name)

        for var in to_delete:
            if var in values:
                del values[var]

        return values

    @classmethod
    def get(cls, session: ManagerSession, name: str) -> FeatureTemplate:
        """Gets feature template model corresponding to existing feature template based on provided name

        Args:
            session: ManagerSession
            name: name of the existing feature template

        Returns:
            FeatureTemplate: filed out feature template model
        """
        from catalystwan.utils.feature_template.choose_model import choose_model

        template_info = (
            session.api.templates._get_feature_templates(summary=False).filter(name=name).single_or_default()
        )
        template_definition_as_dict = json.loads(cast(str, template_info.template_definiton))

        feature_template_model = choose_model(type_value=template_info.template_type)

        values_from_template_definition = find_template_values(template_definition_as_dict)
        flattened_values = flatten_dict(values_from_template_definition)

        return feature_template_model(
            template_name=template_info.name,
            template_description=template_info.description,
            device_models=[DeviceModel(model) for model in template_info.device_type],
            **flattened_values,
        )
